###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import logging

import biolib.seq_io as seq_io
from biolib.external.fasttree import FastTree
from biolib.parallel import Parallel
from biolib.bootstrap import bootstrap_alignment, bootstrap_support
from biolib.common import remove_extension, make_sure_path_exists


class Bootstrap(object):
    """Assess robustness of genome tree by bootstrapping multiple sequence alignment."""

    def __init__(self, cpus):
        """Initialization.

        Parameters
        ----------
        cpus : int
          Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.cpus = cpus

    def _producer(self, replicated_num):
        """Infer tree from bootstrapped multiple sequence alignment.

        Parameters
        ----------
        replicated_num : int
          Unique replicate number.
        """

        output_msa = os.path.join(self.replicate_dir, 'bootstrap_msa.r_' + str(replicated_num) + '.fna')
        bootstrap_alignment(self.msa, output_msa, frac=self.frac)

        fast_tree = FastTree(multithreaded=False)
        output_tree = os.path.join(self.replicate_dir, 'bootstrap_tree.r_' + str(replicated_num) + '.tree')
        fast_tree_output = os.path.join(self.replicate_dir, 'bootstrap_fasttree.r_' + str(replicated_num) + '.out')
        fast_tree.run(output_msa, self.base_type, self.model, output_tree, fast_tree_output)

        return True

    def _progress(self, processed_items, total_items):
        """Report progress of replicates."""

        return '    Processed %d of %d replicates.' % (processed_items, total_items)

    def run(self, input_tree, msa_file, num_replicates, model, base_type, frac, output_dir):
        """Bootstrap multiple sequence alignment.

        Parameters
        ----------
        input_tree : str
          Tree inferred with all data.
        msa_file : str
          File containing multiple sequence alignment for all taxa.
        num_replicates : int
          Number of replicates to perform.
        model : str
          Desired model of evolution.
        base_type : str
          Indicates if bases are nucleotides or amino acids.
        frac : float
          Fraction of alignment to subsample.
        output_dir : str
          Directory for bootstrap trees.
        """

        assert(model in ['wag', 'jtt'])
        assert(base_type in ['nt', 'prot'])

        self.model = model
        self.base_type = base_type
        self.frac = frac

        self.replicate_dir = os.path.join(output_dir, 'replicates')
        make_sure_path_exists(self.replicate_dir)

        # read full multiple sequence alignment
        self.msa = seq_io.read(msa_file)

        # calculate replicates
        self.logger.info('Calculating bootstrap replicates:')
        parallel = Parallel(self.cpus)
        parallel.run(self._producer, None, xrange(num_replicates), self._progress)

        # calculate support values
        rep_tree_files = []
        for rep_index in xrange(num_replicates):
            rep_tree_files.append(os.path.join(self.replicate_dir, 'bootstrap_tree.r_' + str(rep_index) + '.tree'))

        output_tree = os.path.join(output_dir, remove_extension(input_tree) + '.bootstrap.tree')
        bootstrap_support(input_tree, rep_tree_files, output_tree)

        return output_tree



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import logging
import random
from math import floor

import biolib.seq_io as seq_io
from biolib.external.fasttree import FastTree
from biolib.parallel import Parallel
from biolib.bootstrap import bootstrap_support
from biolib.common import remove_extension, make_sure_path_exists

from genometreetk.tree_support import TreeSupport


class JackknifeTaxa(object):
    """Assess robustness by jackknifing taxa in alignment."""

    def __init__(self, cpus):
        """Initialization.

        Parameters
        ----------
        cpus : int
          Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.cpus = cpus

    def _producer(self, replicated_num):
        """Infer tree from jackknifed alignments.

        Parameters
        ----------
        replicated_num : int
          Unique replicate number.
        """

        output_msa = os.path.join(self.replicate_dir, 'jk_taxa.msa.' + str(replicated_num) + '.fna')
        self.jackknife_taxa(self.msa, self.perc_taxa_to_keep, self.outgroup_ids, output_msa)

        fast_tree = FastTree(multithreaded=False)
        output_tree = os.path.join(self.replicate_dir, 'jk_taxa.tree.' + str(replicated_num) + '.tre')
        fast_tree_output = os.path.join(self.replicate_dir, 'jk_taxa.fasttree.' + str(replicated_num) + '.out')
        fast_tree.run(output_msa, 'prot', self.model, output_tree, fast_tree_output)

        return True

    def _progress(self, processed_items, total_items):
        """Report progress of replicates."""

        return '    Processed %d of %d replicates.' % (processed_items, total_items)

    def jackknife_taxa(self, msa, perc_taxa_to_keep, outgroup_ids, output_file):
        """Jackknife alignment to a subset of taxa.

        Parameters
        ----------
        msa : d[seq_id] -> seq
          Full multiple sequence alignment.
        perc_taxa_to_keep : float
          Percentage of marker genes to keep in each replicate.
        outgroup_ids : set
          Labels of outgroup taxa.
        output_file : str
          File to write bootstrapped alignment.
        """

        # randomly select ingroup taxa
        ingroup_taxa = set(msa.keys()) - outgroup_ids
        taxa_to_keep = random.sample(ingroup_taxa, int(floor(len(ingroup_taxa) * perc_taxa_to_keep)))

        taxa_to_keep = set(taxa_to_keep).union(outgroup_ids)

        fout = open(output_file, 'w')
        for seq_id, seq in msa.iteritems():
            if seq_id in taxa_to_keep:
                fout.write('>' + seq_id + '\n')
                fout.write(seq + '\n')
        fout.close()

    def run(self, input_tree, msa_file, outgroup_file, perc_taxa_to_keep, num_replicates, model, output_dir):
        """Jackknife taxa.

        Parameters
        ----------
        input_tree : str
          Tree inferred with all data.
        msa_file : str
          File containing multiple sequence alignment for all taxa.
        outgroup_file : str
          File indicating labels of outgroup taxa.
        perc_taxa_to_keep : float
          Percentage of taxa to keep in each replicate.
        num_replicates : int
          Number of replicates to perform.
        model : str
          Desired model of evolution.
        output_dir : str
          input_tree directory for bootstrap trees.
        """

        assert(model in ['wag', 'jtt'])

        self.perc_taxa_to_keep = perc_taxa_to_keep
        self.model = model
        self.replicate_dir = os.path.join(output_dir, 'replicates')
        make_sure_path_exists(self.replicate_dir)

        # read outgroup taxa
        self.outgroup_ids = set()
        if outgroup_file:
            for line in open(outgroup_file):
                self.outgroup_ids.add(line.strip())

        # read full multiple sequence alignment
        self.msa = seq_io.read(msa_file)

        # calculate replicates
        self.logger.info('')
        self.logger.info('  Calculating jackknife taxa replicates:')
        #parallel = Parallel(self.cpus)
        #parallel.run(self._producer, None, xrange(num_replicates), self._progress)

        # calculate support
        rep_tree_files = []
        for rep_index in xrange(num_replicates):
            rep_tree_files.append(os.path.join(self.replicate_dir, 'jk_taxa.tree.' + str(rep_index) + '.tre'))

        tree_support = TreeSupport()
        output_tree = os.path.join(output_dir, remove_extension(input_tree) + '.jk_taxa.tree')
        tree_support.subset_taxa(input_tree, rep_tree_files, output_tree)

        return output_tree


###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging

import dendropy


class CombineSupport(object):
    """Combine all support values into a single tree."""

    def __init__(self):
        """Initialization."""
        self.logger = logging.getLogger()

    def _collect_support_values(self, tree):
        """Get support value for each node in preorder traversal.

        Parameters
        ----------
        tree : dendropy.Tree
          Tree to obtain support values from.
        """

        preorder_support = []
        for node in tree.preorder_node_iter():
            if node.is_internal():
                preorder_support.append(node.label)

        return preorder_support

    def run(self, support_type, bootstrap_tree, jk_marker_tree, jk_taxa_tree, output_tree):
        """Create new tree indicating combined support values.

        Tree can either be decorated with the average support value
        or the minimum support value as determine by support_type.

        Parameters
        ----------
        support_type : str => 'average' or 'minimum'
            Type of support value to calculate.
        bootstrap_tree : str
          Tree with bootstrap support values.
        jk_marker_tree : str
          Tree with jackknife marker support values.
        jk_taxa_tree : str
          Tree with jackknife taxa support values.
        output_file : str
          File to write tree with combined support values.
        """

        assert(support_type in ['average', 'minimum'])

        tree = dendropy.Tree.get_from_path(bootstrap_tree, schema='newick', rooting='force-rooted', preserve_underscores=True)
        bootstrap_support = self._collect_support_values(tree)

        tree = dendropy.Tree.get_from_path(jk_marker_tree, schema='newick', rooting='force-rooted', preserve_underscores=True)
        jk_marker_support = self._collect_support_values(tree)

        tree = dendropy.Tree.get_from_path(jk_taxa_tree, schema='newick', rooting='force-rooted', preserve_underscores=True)
        jk_taxa_support = self._collect_support_values(tree)

        internalNodeNum = 0
        for node in tree.preorder_node_iter():
            if node.is_internal():
                if support_type == 'average':
                    support = (int(bootstrap_support[internalNodeNum]) + int(jk_marker_support[internalNodeNum]) + int(jk_taxa_support[internalNodeNum])) / 3.0
                elif support_type == 'minimum':
                    support = min(int(bootstrap_support[internalNodeNum]), int(jk_marker_support[internalNodeNum]), int(jk_taxa_support[internalNodeNum]))

                node.label = '%s' % str(int(support + 0.5))
                internalNodeNum += 1

        tree.write_to_path(output_tree, schema='newick', suppress_rooting=True, unquoted_underscores=True)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging

from biolib.external.fasttree import FastTree
from biolib.common import make_sure_path_exists

from genometreetk.default_values import DefaultValues
from genometreetk.common import create_concatenated_alignment
from genometreetk.jackknife_markers import JackknifeMarkers

import dendropy


class LgtTest(object):
    """Identify gene trees that may have undergone one or more lateral transfer.

    Specifically, the following test is applied:
      1) infer a jackknifed genome tree by randomly subsampling marker genes under 100 replicates
      2) identify all splits with at least a minimum jackknife support value, and
         where at least a certain percentage of the taxa fall on each side of the split
      3) determine how many of these "well-support, internal" splits are recovered in each gene tree
      4) filter gene trees which do not recover a specific percentage of these splits
    """

    def __init__(self, cpus):
        """Initialize.

        Parameters
        ----------
        cpus : int
            Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.cpus = cpus

    def run(self, genome_ids,
                    marker_genes,
                    hmm_model_file,
                    min_support,
                    min_per_taxa,
                    perc_markers_to_jackknife,
                    gene_tree_dir,
                    alignment_dir,
                    output_dir):
        """Identify gene trees which do not recover well-support, internal splits in a jackknifed genome tree.

        Parameters
        ----------
        genome_ids : iterable
            Genomes of interest.
        marker_genes : iterable
            Unique ids of marker genes.
        hmm_model_file : str
            File containing HMMs for each marker gene.
        min_support : float
            Minimum jackknife support of splits to use during LGT filtering [0, 1].
        min_per_taxa : float
            Minimum percentage of taxa required to consider a split during LGT filtering [0, 1].
        perc_markers_to_jackknife : float
            Percentage of taxa to keep during marker jackknifing [0, 1].
        gene_tree_dir : str
            Directory containing gene trees.
        alignment_dir : str
            Directory containing multiple sequence alignments.
        output_dir : str
            Output directory.
        """

        output_dir = os.path.join(output_dir, 'jackknife_markers')
        make_sure_path_exists(output_dir)

        # create concatenated alignment file
        self.logger.info('Concatenating alignments.')
        concatenated_alignment_file = os.path.join(output_dir, 'concatenated_alignment.faa')
        marker_file = os.path.join(output_dir, 'concatenated_markers.tsv')
        create_concatenated_alignment(genome_ids, marker_genes, alignment_dir, concatenated_alignment_file, marker_file)

        # create concatenated genome tree
        self.logger.info('Inferring concatenated genome tree.')
        concatenated_tree = os.path.join(output_dir, 'concatenated.tree')
        concatenated_tree_log = os.path.join(output_dir, 'concatenated.tree.log')
        log_file = os.path.join(output_dir, 'concatenated.fasttree.log')
        fast_tree = FastTree(multithreaded=True)
        fast_tree.run(concatenated_alignment_file, 'prot', 'wag', concatenated_tree, concatenated_tree_log, log_file)

        # calculate jackknife support values
        self.logger.info('Calculating jackknife marker support values.')
        jackknife_markers = JackknifeMarkers(self.cpus)
        jackknife_tree = jackknife_markers.run(concatenated_tree, concatenated_alignment_file, marker_file, perc_markers_to_jackknife, 100, 'wag', output_dir)
        # jackknife_tree = os.path.join(output_dir, 'concatenated.jk_markers.tree')

        # identify well-support, internal splits
        self.logger.info('Identifying well-support, internal splits.')
        tree = dendropy.Tree.get_from_path(jackknife_tree, schema='newick', rooting='force-unrooted', preserve_underscores=True)
        num_leaves = len(tree.leaf_nodes())

        num_internal_nodes = 0
        num_major_splits = 0
        well_supported_major_splits = 0

        splits = []
        for node in tree.internal_nodes():
            num_internal_nodes += 1

            num_node_leaves = len(node.leaf_nodes())
            if min(num_node_leaves, num_leaves - num_node_leaves) >= max(min_per_taxa * num_leaves, 2):
                num_major_splits += 1

                if int(node.label) > (min_support * 100.0):
                    well_supported_major_splits += 1
                    split = set([x.taxon.label for x in node.leaf_nodes()])
                    splits.append((split, node.edge_length))

        self.logger.info('# internal nodes: %d' % num_internal_nodes)
        self.logger.info('# major splits: %d' % num_major_splits)
        self.logger.info('# well-supported, major splits: %d' % well_supported_major_splits)

        # filter gene trees that do not recover well-support, internal splits
        self.logger.info('Filtering gene trees.')

        distances = {}
        for i, mg in enumerate(sorted(marker_genes)):
            sys.stdout.write('==> Processed %d of %d (%.2f) gene trees.\r' % (i + 1, len(marker_genes), (i + 1) * 100.0 / len(marker_genes)))
            sys.stdout.flush()

            # read gene tree
            f = mg + '.tree'
            gene_tree_file = os.path.join(gene_tree_dir, f)
            gene_tree = dendropy.Tree.get_from_path(gene_tree_file, schema='newick', rooting='force-unrooted', preserve_underscores=True)

            # prune gene tree so each genome is present exactly once
            processed_genome_ids = set()
            taxa_to_prune = []
            for node in gene_tree.leaf_nodes():
                genome_id = node.taxon.label.split(DefaultValues.SEQ_CONCAT_CHAR)[0]

                if genome_id in processed_genome_ids or genome_id not in genome_ids:
                    taxa_to_prune.append(node.taxon)

                processed_genome_ids.add(genome_id)

            gene_tree.prune_taxa(taxa_to_prune)

            # rename nodes to contain only genome id
            gene_tree_taxa_set = set()
            for node in gene_tree.leaf_nodes():
                genome_id = node.taxon.label.split(DefaultValues.SEQ_CONCAT_CHAR)[0]
                node.taxon.label = genome_id
                gene_tree_taxa_set.add(genome_id)

            # re-encode the split system over the new taxon namespace
            gene_tree.migrate_taxon_namespace(dendropy.TaxonNamespace(gene_tree_taxa_set))
            gene_tree.encode_bipartitions()
            split_bitmasks = set(b.split_bitmask for b in gene_tree.bipartition_encoding)

            # determine number of splits recovered by or compatible with this gene tree
            recovered_splits = 0
            compatible_splits = 0
            compatible_edge_length = 0
            for split, edge_length in splits:
                common_taxa_labels = split.intersection(gene_tree_taxa_set)

                common_split = gene_tree.taxon_namespace.taxa_bitmask(labels=common_taxa_labels)
                normalized_split = dendropy.Bipartition.normalize_bitmask(
                                    bitmask=common_split,
                                    fill_bitmask=gene_tree.taxon_namespace.all_taxa_bitmask(),
                                    lowest_relevant_bit=1)

                if normalized_split in split_bitmasks:
                    recovered_splits += 1

                if gene_tree.is_compatible_with_bipartition(dendropy.Bipartition(bitmask=normalized_split, is_rooted=False)):
                    compatible_splits += 1
                    compatible_edge_length += edge_length

            perc_recovered_splits = recovered_splits * 100.0 / len(splits)
            perc_comp_splits = compatible_splits * 100.0 / len(splits)
            norm_comp_edge_length = float(compatible_edge_length) / sum([s[1] for s in splits])

            # calculate weighted Robinson-Foulds (Manhattan) and Felsenstein's Euclidean
            # distances to the concatenated genome tree
            pruned_tree = tree.clone(depth=2)
            pruned_tree.retain_taxa_with_labels(gene_tree.taxon_namespace.labels())
            pruned_tree.migrate_taxon_namespace(gene_tree.taxon_namespace)
            pruned_tree.encode_bipartitions()

            pruned_tree_edge_len = sum([e.length for e in pruned_tree.edges() if e.length])
            gene_tree_edge_len = sum([e.length for e in gene_tree.edges() if e.length])
            pruned_tree.scale_edges(1.0 / pruned_tree_edge_len)
            gene_tree.scale_edges(1.0 / gene_tree_edge_len)

            manhattan = dendropy.calculate.treecompare.weighted_robinson_foulds_distance(pruned_tree, gene_tree)
            euclidean = dendropy.calculate.treecompare.euclidean_distance(pruned_tree, gene_tree)

            distances[mg] = (perc_recovered_splits, perc_comp_splits, norm_comp_edge_length, manhattan, euclidean)

        return distances, num_internal_nodes, num_major_splits, well_supported_major_splits



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import operator
from collections import defaultdict
import multiprocessing as mp

import biolib.seq_io as seq_io

from genometreetk.exceptions import GenomeTreeTkError
from genometreetk.common import (read_gtdb_taxonomy,
                                 read_gtdb_ncbi_taxonomy,
                                 read_gtdb_ncbi_organism_name,
                                 species_label)
from genometreetk.aai import mismatches


class Cluster(object):
    """Cluster genomes based on AAI of concatenated alignment.

    Genomes are preferentially assigned to representatives from
    RefSeq, followed by GenBank, and final User genomes. This
    is done to ensure the majority of genomes are assigned to
    publicly available representatives. However, this means
    a genome will not necessarily be assigned to the representative
    with the highest AAI. Furthermore, genomes are not clustered
    if they have different valid species names. NCBI genomes are
    never assigned to User representatives to ensure they always
    appear in trees without User genomes.
    """

    def __init__(self, cpus):
        """Initialization.

        Parameters
        ----------
        cpus : int
            Maximum number of cpus/threads to use.
        """

        self.logger = logging.getLogger()

        self.cpus = cpus

        self.source_order = {'R': 0,  # RefSeq
                                'G': 1,  # GenBank
                                'U': 2}  # User

    def _reassign_representative(self,
                                    cur_representative_id,
                                    cur_max_mismatches,
                                    new_representative_id,
                                    new_max_mismatches,
                                    trusted_user_genomes):
        """Determines best representative genome.

        Genomes are preferentially assigned to representatives based on
        source repository (RefSeq => GenBank => Trusted User => User) and AAI.
        """

        if not cur_representative_id:
            # no currently assigned representative
            return new_representative_id, new_max_mismatches

        cur_source = self.source_order[cur_representative_id[0]]
        new_source = self.source_order[new_representative_id[0]]
        if new_representative_id in trusted_user_genomes:
            new_source = self.source_order['G']
        
        if new_source < cur_source:
            # give preference to genome source
            return new_representative_id, new_max_mismatches
        elif new_source == cur_source:
            # for the same source, find the representative with the highest AAI
            if new_max_mismatches < cur_max_mismatches:
                return new_representative_id, new_max_mismatches

        return cur_representative_id, cur_max_mismatches

    def __worker(self,
                 representatives,
                 bac_seqs,
                 ar_seqs,
                 aai_threshold,
                 metadata_file,
                 trusted_user_genomes,
                 queue_in,
                 queue_out):
        """Process genomes in parallel."""

        # determine genus of each genome and representatives belonging to a genus
        gtdb_taxonomy = read_gtdb_taxonomy(metadata_file)
        ncbi_taxonomy = read_gtdb_ncbi_taxonomy(metadata_file)
        ncbi_organism_names = read_gtdb_ncbi_organism_name(metadata_file)
        species = species_label(gtdb_taxonomy, ncbi_taxonomy, ncbi_organism_names)

        # determine genus of all genomes and representatives
        genus = {}
        reps_from_genus = defaultdict(set)
        for genome_id, t in gtdb_taxonomy.iteritems():
            g = None
            if len(t) >= 6 and t[5] != 'g__':
                g = t[5]
            elif genome_id in species:
                g = species[genome_id].split(' ')[0][3:]

            if g:
                genus[genome_id] = g

                if genome_id in representatives:
                    reps_from_genus[g].add(genome_id)

        while True:
            genome_id = queue_in.get(block=True, timeout=None)
            if genome_id == None:
                break

            genome_genus = genus.get(genome_id, None)
            genome_species = species.get(genome_id, None)
            genome_bac_seq = bac_seqs[genome_id]
            genome_ar_seq = ar_seqs[genome_id]

            cur_aai_threshold = aai_threshold
            assigned_representative = None

            bac_max_mismatches = (1.0 - aai_threshold) * (len(genome_bac_seq) - genome_bac_seq.count('-'))
            ar_max_mismatches = (1.0 - aai_threshold) * (len(genome_ar_seq) - genome_ar_seq.count('-'))

            # speed up computation by first comparing genome
            # to representatives of the same genus
            cur_reps_from_genus = reps_from_genus.get(genome_genus, set())
            for rep_id in cur_reps_from_genus:
                # do not cluster genomes from different named species
                rep_species = species.get(rep_id, None)
                if rep_species and genome_species and rep_species != genome_species:
                    continue

                # do not cluster NCBI or trusted User genomes with User representatives
                if rep_id.startswith('U_') and (not genome_id.startswith('U__') or genome_id in trusted_user_genomes):
                    continue

                rep_bac_seq = bac_seqs[rep_id]
                rep_ar_seq = ar_seqs[rep_id]

                m = mismatches(rep_bac_seq, genome_bac_seq, bac_max_mismatches)
                if m is not None:  # necessary to distinguish None and 0
                    assigned_representative, bac_max_mismatches = self._reassign_representative(assigned_representative,
                                                                                                bac_max_mismatches,
                                                                                                rep_id,
                                                                                                m,
                                                                                                trusted_user_genomes)
                else:
                    m = mismatches(rep_ar_seq, genome_ar_seq, ar_max_mismatches)
                    if m is not None:  # necessary to distinguish None and 0
                        assigned_representative, ar_max_mismatches = self._reassign_representative(assigned_representative,
                                                                                                    ar_max_mismatches,
                                                                                                    rep_id,
                                                                                                    m,
                                                                                                    trusted_user_genomes)

            # compare genome to remaining representatives
            remaining_reps = representatives.difference(cur_reps_from_genus)
            for rep_id in remaining_reps:
                # do not cluster genomes from different named species
                rep_species = species.get(rep_id, None)
                if rep_species and genome_species and rep_species != genome_species:
                    continue

                # do not cluster genomes from different named groups
                skip = False
                for rep_taxon, taxon in zip(gtdb_taxonomy[rep_id], gtdb_taxonomy[genome_id]):
                    rep_taxon = rep_taxon[3:]
                    taxon = taxon[3:]
                    if rep_taxon and taxon and rep_taxon != taxon:
                        skip = True
                        break
                if skip:
                    continue

                # do not cluster NCBI or trusted User genomes with User representatives
                if rep_id.startswith('U_') and (not genome_id.startswith('U__') or genome_id in trusted_user_genomes):
                    continue

                rep_bac_seq = bac_seqs[rep_id]
                rep_ar_seq = ar_seqs[rep_id]

                m = mismatches(rep_bac_seq, genome_bac_seq, bac_max_mismatches)
                if m is not None:  # necessary to distinguish None and 0
                    assigned_representative, bac_max_mismatches = self._reassign_representative(assigned_representative,
                                                                                                bac_max_mismatches,
                                                                                                rep_id,
                                                                                                m,
                                                                                                trusted_user_genomes)
                else:
                    m = mismatches(rep_ar_seq, genome_ar_seq, ar_max_mismatches)
                    if m is not None:  # necessary to distinguish None and 0
                        assigned_representative, ar_max_mismatches = self._reassign_representative(assigned_representative,
                                                                                                    ar_max_mismatches,
                                                                                                    rep_id,
                                                                                                    m,
                                                                                                    trusted_user_genomes)

            queue_out.put((genome_id, assigned_representative))

    def __writer(self, representatives, num_genomes, output_file, writer_queue):
        """Process representative assignments from each process.

        Parameters
        ----------
        representatives : set
            Initial set of representative genomes.
        num_genomes : int
            Number of genomes being processed.
        output_file : str
            Output file specifying genome clustering.
        """

        # initialize clusters
        clusters = {}
        for rep_id in representatives:
            clusters[rep_id] = []

        # gather results for each genome
        processed_genomes = 0
        while True:
            genome_id, assigned_representative = writer_queue.get(block=True, timeout=None)
            if genome_id == None:
              break

            processed_genomes += 1
            statusStr = 'Finished processing %d of %d (%.2f%%) genomes.' % (processed_genomes,
                                                                            num_genomes,
                                                                            float(processed_genomes) * 100 / num_genomes)
            sys.stdout.write('%s\r' % statusStr)
            sys.stdout.flush()

            if assigned_representative:
                clusters[assigned_representative].append(genome_id)

        sys.stdout.write('\n')

        # write out cluster
        fout = open(output_file, 'w')
        clustered_genomes = 0
        for c, cluster_rep in enumerate(sorted(clusters, key=lambda x: len(clusters[x]), reverse=True)):
            cluster_str = 'cluster_%d' % (c + 1)
            cluster = clusters[cluster_rep]
            clustered_genomes += len(cluster)
            fout.write('%s\t%s\t%d\t%s\n' % (cluster_rep, cluster_str, len(cluster) + 1, ','.join(cluster)))

        fout.close()

        self.logger.info('Assigned %d genomes to representatives.' % clustered_genomes)

    def _cluster(self,
                    representatives,
                    genomes_to_process,
                    aai_threshold,
                    ar_seqs,
                    bac_seqs,
                    metadata_file,
                    trusted_user_genomes,
                    output_file):
        """Assign genomes to representatives.


        Parameters
        ----------
        representatives : set
            Initial set of representative genomes.
        genomes_to_process : list
            Genomes to process for identification of new representatives.
        aai_threshold : float
              AAI threshold for assigning a genome to a representative.
        ar_seqs : d[genome_id] -> alignment
            Alignment of archaeal marker genes.
        bac_seqs : d[genome_id] -> alignment
            Alignment of bacterial marker genes.
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        trusted_user_genomes : set
            Trusted User genomes to treat as if they were in GenBank.
        output_file : str
            Output file specifying genome clustering.
        """

        # populate worker queue with data to process
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for genome_id in genomes_to_process:
          worker_queue.put(genome_id)

        for _ in range(self.cpus):
          worker_queue.put(None)

        try:
          worker_proc = [mp.Process(target=self.__worker, args=(representatives,
                                                                    bac_seqs,
                                                                    ar_seqs,
                                                                    aai_threshold,
                                                                    metadata_file,
                                                                    trusted_user_genomes,
                                                                    worker_queue,
                                                                    writer_queue)) for _ in range(self.cpus)]
          write_proc = mp.Process(target=self.__writer, args=(representatives,
                                                              len(genomes_to_process),
                                                              output_file,
                                                              writer_queue))

          write_proc.start()

          for p in worker_proc:
              p.start()

          for p in worker_proc:
              p.join()

          writer_queue.put((None, None))
          write_proc.join()
        except:
          for p in worker_proc:
            p.terminate()

          write_proc.terminate()

    def run(self,
            representatives_file,
            trusted_user_file,
            ar_msa_file,
            bac_msa_file,
            aai_threshold,
            metadata_file,
            output_file):
        """Identify additional representatives based on AAI between aligned sequences.

        Parameters
        ----------
        representatives_file : str
            File listing genome identifiers as initial representatives.
        trusted_user_file : str
            File listing trusted User genomes that should be treated as if they are in GenBank.
        ar_msa_file : str
            Name of file containing canonical archaeal multiple sequence alignment.
        bac_msa_file : str
            Name of file containing canonical bacterial multiple sequence alignment.
        aai_threshold : float
              AAI threshold for clustering genomes to a representative.
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        output_file : str
            Output file specifying genome clustering.
        """

        # read sequences
        ar_seqs = seq_io.read_fasta(ar_msa_file)
        bac_seqs = seq_io.read_fasta(bac_msa_file)
        self.logger.info('Identified %d archaeal sequences in MSA.' % len(ar_seqs))
        self.logger.info('Identified %d bacterial sequences in MSA.' % len(bac_seqs))

        if len(ar_seqs) != len(bac_seqs):
            self.logger.error('Archaeal and bacterial MSA files do not contain the same number of sequences.')
            raise GenomeTreeTkError('Error with MSA input files.')

        genome_to_consider = set(ar_seqs.keys())

        # read initial representatives
        rep_genomes = set()
        for line in open(representatives_file):
            if line[0] == '#':
                continue

            genome_id = line.rstrip().split('\t')[0]
            if genome_id not in genome_to_consider:
                self.logger.error('Representative genome %s has no sequence data.' % genome_id)
            rep_genomes.add(genome_id)

        self.logger.info('Identified %d representatives.' % len(rep_genomes))
        
        # read trusted User genomes
        trusted_user_genomes = set()
        if trusted_user_file:
            for line in open(trusted_user_file):
                line_split = line.split('\t')
                trusted_user_genomes.add(line_split[0])
                
        self.logger.info('Identified %d trusted User genomes.' % len(trusted_user_genomes)) 

        # cluster genomes to representatives
        genomes_to_cluster = genome_to_consider - rep_genomes
        self.logger.info('Comparing %d genomes to %d representatives with threshold = %.3f.' % (len(genomes_to_cluster),
                                                                                                len(rep_genomes),
                                                                                                aai_threshold))
        self._cluster(rep_genomes,
                                    genomes_to_cluster,
                                    aai_threshold,
                                    ar_seqs,
                                    bac_seqs,
                                    metadata_file,
                                    trusted_user_genomes,
                                    output_file)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import logging
import random
from math import floor

import biolib.seq_io as seq_io
from biolib.external.fasttree import FastTree
from biolib.parallel import Parallel
from biolib.common import remove_extension, make_sure_path_exists
from biolib.bootstrap import bootstrap_support


class JackknifeMarkers(object):
    """Assess robustness by jackkifing genes in alignment."""

    def __init__(self, cpus):
        """Initialization.

        Parameters
        ----------
        cpus : int
          Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.cpus = cpus

    def _producer(self, replicated_num):
        """Infer tree from jackknifed alignments.

        Parameters
        ----------
        replicated_num : int
          Unique replicate number.
        """

        output_msa = os.path.join(self.replicate_dir, 'jk_markers.msa.' + str(replicated_num) + '.faa')
        self.jackknife_alignment(self.msa, self.perc_markers_to_keep, self.marker_lengths, output_msa)

        fast_tree = FastTree(multithreaded=False)
        output_tree = os.path.join(self.replicate_dir, 'jk_markers.tree.' + str(replicated_num) + '.tre')
        fast_tree_output = os.path.join(self.replicate_dir, 'jk_markers.fasttree.' + str(replicated_num) + '.out')
        fast_tree.run(output_msa, 'prot', self.model, output_tree, fast_tree_output)

        return True

    def _progress(self, processed_items, total_items):
        """Report progress of replicates."""

        return '==> Processed %d of %d replicates.' % (processed_items, total_items)

    def jackknife_alignment(self, msa, perc_markers_to_keep, marker_lengths, output_file):
        """Jackknife alignment to a subset of marker genes.

        The marker_lengths must be specified in the order
        in which genes were concatenated.

        Parameters
        ----------
        msa : d[seq_id] -> seq
          Full multiple sequence alignment.
        perc_markers_to_keep : float
          Percentage of marker genes to keep in each replicate [0, 1].
        marker_lengths : list
          Length of each marker gene.
        output_file : str
          File to write bootstrapped alignment.
        """
        markers_to_keep = random.sample(xrange(0, len(marker_lengths)), int(floor(perc_markers_to_keep * len(marker_lengths))))

        start_pos = [0]
        for index, ml in enumerate(marker_lengths):
            start_pos.append(start_pos[index] + ml)

        mask = [0] * sum(marker_lengths)
        for marker_index in markers_to_keep:
            start = start_pos[marker_index]
            end = start + marker_lengths[marker_index]
            mask[start:end] = [1] * (end - start)

        fout = open(output_file, 'w')
        for seq_id, seq in msa.iteritems():
            fout.write('>' + seq_id + '\n')
            sub_seq = ''.join([base for base, m in zip(seq, mask) if m == 1])
            fout.write(sub_seq + '\n')
        fout.close()

    def run(self, input_tree, msa_file, marker_file, perc_markers_to_keep, num_replicates, model, output_dir):
        """Jackknife marker genes.

        Marker file should have the format:
          <marker id>\t<marker name>\t<marker desc>\t<length>\n

        Parameters
        ----------
        input_tree : str
          Tree inferred with all data.
        msa_file : str
          File containing multiple sequence alignment for all taxa.
        marker_file : str
          File indicating database id, HMM name, description and length of each marker in the alignment.
        perc_markers_to_keep : float [0, 1]
          Percentage of marker genes to keep in each replicate.
        num_replicates : int
          Number of replicates to perform.
        model : str
          Desired model of evolution.
        output_dir : str
          Output directory for jackkife trees.
        """

        assert(model in ['wag', 'jtt'])

        self.model = model
        self.perc_markers_to_keep = perc_markers_to_keep
        self.replicate_dir = os.path.join(output_dir, 'replicates')
        make_sure_path_exists(self.replicate_dir)
        # determine length of each marker gene in alignment
        self.marker_lengths = []
        with open(marker_file) as f:
	    f.readline()
	    for line in f:
            	line_split = line.split('\t')
            	self.marker_lengths.append(int(line_split[3]))

        # read full multiple sequence alignment
        self.msa = seq_io.read(msa_file)

        # calculate replicates
        self.logger.info('Calculating jackknife marker replicates:')
        parallel = Parallel(self.cpus)
        parallel.run(self._producer, None, xrange(num_replicates), self._progress)

        # calculate support
        rep_tree_files = []
        for rep_index in xrange(num_replicates):
            rep_tree_files.append(os.path.join(self.replicate_dir, 'jk_markers.tree.' + str(rep_index) + '.tre'))

        output_tree = os.path.join(output_dir, remove_extension(input_tree) + '.jk_markers.tree')
        bootstrap_support(input_tree, rep_tree_files, output_tree)

        return output_tree



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import multiprocessing as mp
import logging
from collections import defaultdict

import biolib.seq_io as seq_io
from biolib.external.hmmer import HMMER

from genometreetk.default_values import DefaultValues


class AlignMarkers(object):
    """Align genes to HMM."""

    def __init__(self, cpus):
        """Initialize.

        Parameters
        ----------
        cpus : int
            Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.cpus = cpus

        self.protein_file_ext = DefaultValues.PROTEIN_FILE_EXTENSION
        self.pfam_extension = DefaultValues.PFAM_EXTENSION
        self.tigr_extension = DefaultValues.TIGR_EXTENSION

    def _genes_in_genomes(self, genome_ids, genome_dirs):
        """Get genes within genomes.

        Parameters
        ----------
        genome_ids : iterable
            Genomes of interest.
        genome_dirs : d[assembly_accession] -> directory
            Path to files for individual genomes.

        Returns
        -------
        d[genome_id][family_id] -> [(gene_id_1, bitscore), ..., (gene_id_N, bitscore)]
            Genes within each genome.
        """

        genes_in_genome = {}
        for genome_id in genome_ids:
            genome_dir = genome_dirs[genome_id]

            marker_id_to_gene_id = defaultdict(list)

            assembly = genome_dir[genome_dir.rfind('/') + 1:]
            tophit_file = os.path.join(genome_dir, assembly + self.pfam_extension)
            with open(tophit_file) as f:
                f.readline()
                for line in f:
                    line_split = line.split('\t')

                    gene_id = line_split[0]
                    hits = line_split[1].split(';')
                    for hit in hits:
                        pfam_id, _evalue, bitscore = hit.split(',')
                        marker_id_to_gene_id[pfam_id].append((gene_id, float(bitscore)))

            tophit_file = os.path.join(genome_dir, assembly + self.tigr_extension)
            with open(tophit_file) as f:
                f.readline()
                for line in f:
                    line_split = line.split('\t')

                    gene_id = line_split[0]
                    hits = line_split[1].split(';')
                    for hit in hits:
                        tigrfam_id, _evalue, bitscore = hit.split(',')
                        marker_id_to_gene_id[tigrfam_id].append((gene_id, float(bitscore)))

            genes_in_genome[genome_id] = marker_id_to_gene_id

        return genes_in_genome

    def _run_hmm_align(self, genome_ids,
                                genome_dirs,
                                genes_in_genomes,
                                ignore_multi_copy,
                                output_msa_dir,
                                output_model_dir,
                                queue_in,
                                queue_out):
        """Run each marker gene in a separate thread.

        Only the gene with the highest bitscore is used for genomes with
        multiple hits to a given protein family.

        Parameters
        ----------
        genome_ids : iterable
            Genomes of interest.
        genome_dirs : d[assembly_accession] -> directory
            Path to files for individual genomes.
        genes_in_genomes : d[genome_id][family_id] -> [(gene_id_1, bitscore), ..., (gene_id_N, bitscore)]
            Genes within each genome.
        ignore_multi_copy : bool
            Flag indicating if genes with multiple hits should be ignored (True) or the gene with the highest bitscore taken (False).
        output_msa_dir : str
            Output directory for multiple sequence alignment.
        output_model_dir : str
            Output directory for HMMs.
        queue_in : Queue
            Input queue for parallel processing.
        queue_out : Queue
            Output queue for parallel processing.
        """

        while True:
            marker_id = queue_in.get(block=True, timeout=None)
            if marker_id == None:
                break

            marker_seq_file = os.path.join(output_msa_dir, marker_id + '.faa')
            fout = open(marker_seq_file, 'w')
            for genome_id in genome_ids:
                genome_dir = genome_dirs[genome_id]

                assembly = genome_dir[genome_dir.rfind('/') + 1:]
                genes_file = os.path.join(genome_dir, assembly + self.protein_file_ext)
                seqs = seq_io.read_fasta(genes_file)

                hits = genes_in_genomes[genome_id].get(marker_id, None)
                if not hits or (ignore_multi_copy and len(hits) > 1):
                    continue

                # get gene with highest bitscore
                hits.sort(key=lambda x: x[1], reverse=True)
                gene_id, _bitscore = hits[0]

                fout.write('>' + genome_id + DefaultValues.SEQ_CONCAT_CHAR + gene_id + '\n')
                fout.write(seqs[gene_id] + '\n')
            fout.close()

            hmmer = HMMER('align')
            hmmer.align(os.path.join(output_model_dir, marker_id + '.hmm'), marker_seq_file, os.path.join(output_msa_dir, marker_id + '.aln.faa'), trim=False, outputFormat='Pfam')
            self._mask_alignment(os.path.join(output_msa_dir, marker_id + '.aln.faa'), os.path.join(output_msa_dir, marker_id + '.aln.masked.faa'))

            queue_out.put(marker_id)

    def _report_threads(self, num_genes, writer_queue):
        """Report progress of parallel processing.

        Parameters
        ----------
        num_genes : int
            Number of genes being processed.
        writer_queue : Queue
            Output queue for parallel processing.
        """

        num_processed_genes = 0
        while True:
            marker_id = writer_queue.get(block=True, timeout=None)
            if marker_id == None:
                break

            num_processed_genes += 1
            statusStr = '==> Finished processing %d of %d (%.2f%%) marker genes.' % (num_processed_genes, num_genes, float(num_processed_genes) * 100 / num_genes)
            sys.stdout.write('%s\r' % statusStr)
            sys.stdout.flush()

        sys.stdout.write('\n')

    def _mask_alignment(self, input_file, output_file):
        """Read HMMER alignment in STOCKHOLM format and output masked alignment in FASTA format.

        Parameters
        ----------
        input_file : str
            Input sequence file in STOCKHOLM format.
        output_file : str
            Output sequence file in FASTA format.
        """

        # read STOCKHOLM alignment
        seqs = {}
        for line in open(input_file):
            line = line.rstrip()
            if line == '' or line[0] == '#' or line == '//':
                if 'GC RF' in line:
                    mask = line.split('GC RF')[1].strip()
                continue
            else:
                line_split = line.split()
                seqs[line_split[0]] = line_split[1].upper().replace('.', '-').strip()

        # output masked sequences in FASTA format
        fout = open(output_file, 'w')
        for seq_id, seq in seqs.iteritems():
            fout.write('>' + seq_id + '\n')

            masked_seq = ''.join([seq[i] for i in xrange(0, len(seq)) if mask[i] == 'x'])
            fout.write(masked_seq + '\n')
        fout.close()

    def run(self, genome_ids, genome_dirs, marker_genes, ignore_multi_copy, output_msa_dir, output_model_dir):
        """Perform multithreaded alignment of marker genes using HMM align.

        Parameters
        ----------
        genome_ids : iterable
            Genomes of interest.
        genome_dirs : d[assembly_accession] -> directory
            Path to files for individual genomes.
        marker_genes : iterable
            Unique ids of marker genes to align.
        ignore_multi_copy : bool
            Flag indicating if genes with multiple hits should be ignored (True) or the gene with the highest bitscore taken (False).
        output_msa_dir : str
            Output directory for multiple sequence alignments.
        output_model_dir : str
            Output directory for HMMs.
        """

        # get mapping of marker ids to gene ids for each genome
        self.logger.info('Determining genes in genomes of interest.')
        genes_in_genomes = self._genes_in_genomes(genome_ids, genome_dirs)

        # align marker genes
        self.logger.info('Aligning marker genes:')
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for _, marker_id in enumerate(marker_genes):
            worker_queue.put(marker_id)

        for _ in range(self.cpus):
            worker_queue.put(None)

        try:
            calc_proc = [mp.Process(target=self._run_hmm_align, args=(genome_ids,
                                                                      genome_dirs,
                                                                      genes_in_genomes,
                                                                      ignore_multi_copy,
                                                                      output_msa_dir,
                                                                      output_model_dir,
                                                                      worker_queue,
                                                                      writer_queue)) for _ in range(self.cpus)]
            write_proc = mp.Process(target=self._report_threads, args=(len(marker_genes), writer_queue))

            write_proc.start()

            for p in calc_proc:
                p.start()

            for p in calc_proc:
                p.join()

            writer_queue.put(None)
            write_proc.join()
        except:
            for p in calc_proc:
                p.terminate()

            write_proc.terminate()



###############################################################################
#
# common.py - utility functions used in many places in CheckM
#
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import csv
from collections import defaultdict, namedtuple

import biolib.seq_io as seq_io
from biolib.taxonomy import Taxonomy

from genometreetk.default_values import DefaultValues


def species_label(gtdb_taxonomy, ncbi_taxonomy, ncbi_organism_name):
    """Determine 'best' species label for each genome.

    Genomes preferentially use the GTDB taxonomy,
    NCBI taxonomy, and then NCBI organism name.

    Parameters
    ----------
    gtdb_taxonomy : d[assembly_accession] -> [d__, ..., s__]
        GTDB taxonomy of each genome.
    ncbi_taxonomy : d[assembly_accession] -> [d__, ..., s__]
        NCBI taxonomy of each genome.
    ncbi_organism_name : d[assembly_accession] -> name
        NCBI organism name of each genome.

    Return
    ------
    dict : d[assembly_accession] -> species name
        Species name of each genome.
    """

    taxonomy = Taxonomy()

    species = {}
    species_index = Taxonomy.rank_index['s__']
    for genome_id, taxa in gtdb_taxonomy.iteritems():
        sp = taxa[species_index]
        if sp != 's__':
            species[genome_id] = sp

    for genome_id, taxa in ncbi_taxonomy.iteritems():
        if genome_id in species:
            continue

        sp = taxa[species_index]
        sp = taxonomy.extract_valid_species_name(sp)
        if sp:
            species[genome_id] = sp

    for genome_id, sp in ncbi_organism_name.iteritems():
        if genome_id in species:
            continue

        sp = taxonomy.extract_valid_species_name(sp)
        if sp:
            species[genome_id] = sp

    return species

def read_gtdb_metadata(metadata_file, fields):
    """Parse genome quality from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes in CSV file.
    fields : iterable
        Fields  to read.

    Return
    ------
    dict : d[genome_id] -> namedtuple
        Value for fields indicted by genome IDs.
    """

    gtdb_metadata = namedtuple('gtdb_metadata', ' '.join(fields))
    m = {}

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            headers = row

            genome_index = headers.index('genome')

            indices = []
            for field in fields:
                indices.append(headers.index(field))

            bHeader = False
        else:
            genome_id = row[genome_index]

            values = []
            for i in indices:
                # save values as floats or strings
                try:
                    values.append(float(row[i]))
                except ValueError:
                    values.append(row[i])
            m[genome_id] = gtdb_metadata._make(values)

    return m


def read_gtdb_genome_quality(metadata_file):
    """Parse genome quality from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata, including CheckM estimates, for all genomes.

    Return
    ------
    dict : d[genome_id] -> comp, cont, comp - 4*cont
    """

    genome_quality = {}

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            headers = row
            genome_index = headers.index('genome')
            comp_index = headers.index('checkm_completeness')
            cont_index = headers.index('checkm_contamination')
            bHeader = False
        else:
            genome_id = row[genome_index]
            comp = float(row[comp_index])
            cont = float(row[cont_index])

            genome_quality[genome_id] = [comp, cont, comp - 4*cont]

    return genome_quality


def read_gtdb_phylum(metadata_file):
    """Parse GTDB phylum information from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> phyla
    """

    genome_phyla = {}

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            headers = row
            genome_index = headers.index('genome')
            phylum_index = headers.index('gtdb_phylum')
            bHeader = False
        else:
            genome_id = row[genome_index]
            genome_phyla[genome_id] = row[phylum_index]

    return genome_phyla


def read_gtdb_taxonomy(metadata_file):
    """Parse GTDB taxonomy from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> taxonomy list
    """

    taxonomy = {}

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            headers = row
            genome_index = headers.index('genome')
            taxonomy_index = headers.index('gtdb_taxonomy')
            bHeader = False
        else:
            genome_id = row[genome_index]
            taxa_str = row[taxonomy_index].strip()

            if taxa_str:
                taxonomy[genome_id] = taxa_str.split(';')
            else:
                taxonomy[genome_id] = list(Taxonomy.rank_prefixes)

    return taxonomy


def read_gtdb_ncbi_taxonomy(metadata_file):
    """Parse NCBI taxonomy from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> taxonomy list
    """

    taxonomy = {}

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            headers = row
            genome_index = headers.index('genome')
            taxonomy_index = headers.index('ncbi_taxonomy')
            bHeader = False
        else:
            genome_id = row[genome_index]
            taxa_str = row[taxonomy_index].strip()

            if taxa_str:
                taxonomy[genome_id] = taxa_str.split(';')
            else:
                taxonomy[genome_id] = list(Taxonomy.rank_prefixes)

    return taxonomy


def read_gtdb_ncbi_organism_name(metadata_file):
    """Parse NCBI organism name from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> organism name
        Organism name of each genome.
    """

    d = {}

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            headers = row
            genome_index = headers.index('genome')
            organism_name_index = headers.index('ncbi_organism_name')
            bHeader = False
        else:
            genome_id = row[genome_index]
            organism_name = row[organism_name_index].strip()

            if organism_name:
                d[genome_id] = organism_name

    return d


def read_gtdb_ncbi_type_strain(metadata_file):
    """Parse NCBI type strain from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    set
        Set of genomes marked as type strains by NCBI.
    """

    type_strains = set()

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            headers = row
            genome_index = headers.index('genome')
            type_strain_index = headers.index('ncbi_type_strain')
            bHeader = False
        else:
            genome_id = row[genome_index]
            if bool(row[type_strain_index]):
                type_strains.add(genome_id)

    return type_strains


def read_tree_model(reportFile):
    for line in open(reportFile):
        if 'Model of evolution:' in line:
            modelStr = line[line.find(':') + 1:].strip()

    return modelStr


def read_genome_dir_file(genome_dir_file):
    """Read genome directories from file.

    Parameters
    ----------
    genome_dir_file : str
        File to read.

    Returns
    -------
    dict : d[genome_id] -> directory
        Directory for each genome.
    """

    # read directory for each genome
    genome_dirs = {}
    for line in open(genome_dir_file):
        line_split = line.split('\t')
        genome_dirs[line_split[0]] = line_split[1].strip()

    return genome_dirs


def read_marker_id_file(marker_id_file):
    """Read marker ids from file.

    Parameters
    ----------
    marker_id_file : str
        File to read.

    Returns
    -------
    set
        Marker ids.
    """

    marker_genes = set()
    for line in open(marker_id_file):
        if line[0] == '#':
            continue

        marker_genes.add(line.split('\t')[0].strip())

    return marker_genes


def read_genome_id_file(genome_id_file):
    """Read genome ids from file.

    Parameters
    ----------
    genome_file : str
        File to read.

    Returns
    -------
    set
        NCBI genome ids.
    set
        User genome ids.
    """

    ncbi_genome_ids = set()
    user_genome_ids = set()
    for line in open(genome_id_file):
        if line[0] == '#':
            continue

        if '\t' in line:
            genome_id = line.split('\t')[0].strip()
        else:
            genome_id = line.split()[0].strip()

        if genome_id.startswith('U_'):
            user_genome_ids.add(genome_id)
        else:
            ncbi_genome_ids.add(genome_id)

    return ncbi_genome_ids, user_genome_ids


def create_concatenated_alignment(genome_ids,
                                   marker_genes,
                                   alignment_dir,
                                   concatenated_alignment_file,
                                   marker_file):
    """Create concatenated multiple sequence alignment for all genomes.

    Parameters
    ----------
    genome_ids : iterable
        Genomes of interest.
    marker_genes : iterable
        Unique ids of marker genes.
    alignment_dir : str
        Directory containing multiple sequence alignments.
    concatenated_alignment_file : str
        File to containing concatenated alignment.
    marker_file : str
        File indicating length of each marker in the alignment.
    """

    # Read alignment files. Some genomes may have multiple
    # copies of a marker gene in which case the last one
    # is arbitrarily taken. This is acceptable as all genes
    # are already screen to be conspecific.
    alignments = defaultdict(dict)
    marker_length = {}
    for mg in marker_genes:
        f = mg + '.aln.masked.faa'
        seqs = seq_io.read_fasta(os.path.join(alignment_dir, f))

        for seq_id, seq in seqs.iteritems():
            genome_id = seq_id[0:seq_id.find(DefaultValues.SEQ_CONCAT_CHAR)]

            alignments[mg][genome_id] = seq

            marker_length[mg] = len(seq)

    # create marker file
    fout = open(marker_file, 'w')
    for mg in marker_genes:
        fout.write('%s\t%s\t%s\t%d\n' % (mg, mg, mg, marker_length[mg]))
    fout.close()

    # create concatenated alignment
    concatenated_seqs = {}
    for mg in marker_genes:
        seqs = alignments[mg]

        for genome_id in genome_ids:
            if genome_id in seqs:
                # append alignment
                concatenated_seqs[genome_id] = concatenated_seqs.get(genome_id, '') + seqs[genome_id]
            else:
                # missing gene
                concatenated_seqs[genome_id] = concatenated_seqs.get(genome_id, '') + '-' * marker_length[mg]

    # save concatenated alignment
    seq_io.write_fasta(concatenated_seqs, concatenated_alignment_file)




###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import logging

from biolib.common import make_sure_path_exists
from biolib.external.hmmer import HmmModelParser

from genometreetk.markers.infer_markers import InferMarkers
from genometreetk.markers.lgt_test import LgtTest


class MarkerWorkflow(object):
    """Determine phylogenetically informative marker genes."""

    def __init__(self, genome_dir_file,
                        pfam_model_file,
                        tigrfams_model_dir,
                        cpus):
        """Initialization.

        Parameters
        ----------
        genome_dir_file : str
            File specifying directory for each genome.
        pfam_model_file : str
            File containing Pfam HMMs.
        tigrfams_model_dir : str
            Directory containing TIGRFAMs HMMs.
        cpus : int
            Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.genome_dir_file = genome_dir_file
        self.pfam_model_file = pfam_model_file
        self.tigrfams_model_dir = tigrfams_model_dir

        self.logger.info('Genome dir: %s' % self.genome_dir_file)
        self.logger.info('Pfam models: %s' % self.pfam_model_file)
        self.logger.info('TIGRfam models: %s' % self.tigrfams_model_dir)

        self.cpus = cpus

    def _get_hmms(self, hmms_dir, marker_genes, hmm_model_out, hmm_info_out):
        """Create HMM file with marker genes.

        Parameters
        ----------
        hmms_dir : str
            Directory containing individual HMMs.
        marker_genes : iterable
            Unique ids of marker genes.
        hmm_model_out : str
            File to containing phylogenetically informative HMMs.
        hmm_info_out : str
            File to contain information about HMMs.
        """

        # place all marker genes into a single model file
        fout = open(hmm_model_out, 'w')
        for marker_id in marker_genes:
            for line in open(os.path.join(hmms_dir, marker_id + '.hmm')):
                fout.write(line)
        fout.close()

        # read HMM model metadata
        if hmm_info_out:
            hmm_model_parse = HmmModelParser(hmm_model_out)
            hmm_models = hmm_model_parse.models()
            fout = open(hmm_info_out, 'w')
            fout.write('Model Accession\tName\tDescription\tLength\n')
            for model in hmm_models.values():
                fout.write('%s\t%s\t%s\t%s\n' % (model.acc, model.name, model.desc, model.leng))
            fout.close()

    def run(self, ingroup_file,
            ubiquity,
            single_copy,
            redundancy,
            min_support,
            min_per_taxa,
            perc_markers_to_jackknife,
            restrict_marker_list,
            output_dir):
        """Identify phylogenetic marker genes and save to HMM model file.

        Parameters
        ----------
        ingroup_file : str
            File specifying unique ids of ingroup genomes.
        ubiquity : float
            Threshold for defining ubiquity marker genes.
        single_copy : float
            Threshold for defining a single-copy marker gene.
        redundancy : float
            Threshold for declaring HMMs redundant
        non_conspecific : float
            Non-conspecific threshold for retaining multi-copy gene trees.
        min_support : float
            Minimum jackknife support of splits to use during LGT filtering [0, 1].
        min_per_taxa : float
            Minimum percentage of taxa required to consider a split during LGT filtering [0, 1].
        perc_markers_to_jackknife : float
            Percentage of taxa to keep during marker jackknifing [0, 1].
        restrict_marker_list : str
            Restrict marker set to genes listed within file.
        output_dir : str
            Directory to store results.
        """

        alignment_dir = os.path.join(output_dir, 'alignments')
        make_sure_path_exists(alignment_dir)

        gene_tree_dir = os.path.join(output_dir, 'gene_trees')
        make_sure_path_exists(gene_tree_dir)

        model_dir = os.path.join(output_dir, 'hmm_models')
        make_sure_path_exists(model_dir)

        # identify markers specified in marker restriction list
        valid_marker_genes = set()
        if restrict_marker_list != None:
            for line in open(restrict_marker_list):
                line_split = line.split()

                marker = line_split[0]
                valid_marker_genes.add(marker)

        # identify genes suitable for phylogenetic inference
        self.logger.info('Identifying genes suitable for phylogenetic inference.')
        infer_markers = InferMarkers(self.genome_dir_file,
                                    self.pfam_model_file,
                                    self.tigrfams_model_dir,
                                    self.cpus)
        results = infer_markers.identify_marker_genes(ingroup_file,
                                                        ubiquity, single_copy, redundancy,
                                                        valid_marker_genes,
                                                        alignment_dir, model_dir)
        num_ingroup_genomes, num_ncbi_genome, num_user_genomes, trusted_genome_ids, marker_gene_stats, marker_genes = results
        self.logger.info('Identified %s ubiquitous, single-copy marker genes.' % len(marker_genes))

        # gather all ubiquitous, single-copy HMMs into a single model file
        hmm_model_out = os.path.join(output_dir, 'marker_putative.hmm')
        self.logger.info('Gathering HMMs for putative phylogenetic marker genes HMMs.')
        self._get_hmms(model_dir, marker_genes, hmm_model_out, None)

        # infer gene trees
        self.logger.info('Inferring gene trees.')
        infer_markers.infer_gene_trees(alignment_dir, gene_tree_dir, '.aln.masked.faa')

        # identify gene trees which fail to reproduce the majority
        # of well-supported splits in a jackknifed genome tree
        lgt_test = LgtTest(self.cpus)
        results = lgt_test.run(trusted_genome_ids,
                               marker_genes,
                               hmm_model_out,
                               min_support,
                               min_per_taxa,
                               perc_markers_to_jackknife,
                               gene_tree_dir,
                               alignment_dir,
                               output_dir)
        distances, num_internal_nodes, well_supported_nodes, well_supported_internal_nodes = results

        # write out metadata regarding putative marker genes
        marker_info_out = os.path.join(output_dir, 'marker_putative.tsv')
        fout = open(marker_info_out, 'w')
        fout.write('Model accession\tName\tDescription\tLength\tUbiquity\tSingle copy\tRecovered splits (%)\tCompatible splits (%)\tNormalized compatible split length\tManhattan\tEuclidean\n')

        hmm_model_parse = HmmModelParser(hmm_model_out)
        hmm_models = hmm_model_parse.models()
        for model_acc, model_info in hmm_models.iteritems():
            fout.write('%s\t%s\t%s\t%s\t%.1f\t%.1f\t%.1f\t%.1f\t%.2f\t%.2f\t%.3f\n' % (model_acc,
                                                                               model_info.name,
                                                                               model_info.desc,
                                                                               model_info.leng,
                                                                               marker_gene_stats[model_acc][0],
                                                                               marker_gene_stats[model_acc][1],
                                                                               distances[model_acc][0],
                                                                               distances[model_acc][1],
                                                                               distances[model_acc][2],
                                                                               distances[model_acc][3],
                                                                               distances[model_acc][4]))
        fout.close()

        self.logger.info('Information about marker genes written to: %s.' % marker_info_out)

        # generate summary report
        report_out = os.path.join(output_dir, 'marker_workflow.log')
        fout = open(report_out, 'w')
        fout.write('[markers]\n')
        fout.write('Genome dir: %s\n' % self.genome_dir_file)
        fout.write('Pfam models: %s\n' % self.pfam_model_file)
        fout.write('TIGRfam models: %s\n' % self.tigrfams_model_dir)
        fout.write('')
        fout.write('Number of ingroup genomes: %d\n' % num_ingroup_genomes)
        fout.write('  Number of NCBI genomes: %d\n' % num_ncbi_genome)
        fout.write('  Number of user genomes: %d\n' % num_user_genomes)
        fout.write('Number of trusted ingroup genomes: %d\n' % len(trusted_genome_ids))
        fout.write('')
        fout.write('Ubiquity threshold: %f\n' % ubiquity)
        fout.write('Single-copy threshold: %f\n' % single_copy)
        fout.write('')
        fout.write('LGT test minimum support: %f\n' % min_support)
        fout.write('LGT test minimum percent taxa for internal split: %f\n' % min_per_taxa)
        fout.write('LGT test percentage of markers to jackknife: %f\n' % perc_markers_to_jackknife)
        fout.write('')
        fout.write('Number of internal nodes: %d\n' % num_internal_nodes)
        fout.write('Number of well-supported nodes: %d\n' % well_supported_nodes)
        fout.write('Number of well-supported, internal nodes: %d\n' % well_supported_internal_nodes)
        fout.write('')
        fout.write('Putative marker genes: %d\n' % len(marker_genes))
        fout.close()

        return hmm_model_out



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging

import biolib.seq_io as seq_io
from biolib.misc.time_keeper import TimeKeeper
from biolib.external.fasttree import FastTree
from biolib.external.blast import Blast
from biolib.taxonomy import Taxonomy

import genometreetk.ncbi as ncbi
from genometreetk.common import (read_gtdb_metadata,
                                    read_genome_dir_file,
                                    read_gtdb_taxonomy)


class SSU_Workflow(object):
    """Infer SSU trees."""

    def __init__(self, gtdb_metadata_file, genome_dir_file, cpus):
        """Initialization.

        Parameters
        ----------
        gtdb_metadata_file : str
            File specifying GTDB metadata for each genome.
        genome_dir_file : str
            File specifying directory for each genome.
        cpus : int
            Maximum number of CPUs to use.
        """

        self.logger = logging.getLogger()

        self.gtdb_metadata_file = gtdb_metadata_file
        self.genome_dir_file = genome_dir_file
        self.cpus = cpus

    def _get_ssu_seqs(self,
                      min_ssu_length,
                      min_ssu_contig,
                      gtdb_taxonomy,
                      genomes_to_consider,
                      output_dir):
        """Get 16S sequences from genomes.

        Parameters
        ----------
        min_ssu_length : int
            Minimum required length of 16S sequences.
        min_ssu_contig : int
            Minimum required length of contig containing 16S sequence.
        gtdb_taxonomy : list
            GTDB taxonomy for each genome assembly.
        genomes_to_consider : iterable
            IDs of genomes to obtain 16S sequences.
        output_dir : str
            Desired output directory.

        Returns
        -------
        str
            Name of file with 16S sequences in FASTA format.
        """


        genome_dirs = read_genome_dir_file(self.genome_dir_file)

        ssu_output_file = os.path.join(output_dir, 'gtdb_ssu.fna')

        fout = open(ssu_output_file, 'w')
        no_identified_ssu = 0
        for genome_id in genomes_to_consider:
            ssu_hmm_file = os.path.join(genome_dirs[genome_id], 'ssu_gg_2013_08', 'ssu.hmm_summary.tsv')

            if not os.path.exists(ssu_hmm_file):
                no_identified_ssu += 1
                # self.logger.warning('Genome %s has no identified 16S sequence.' % genome_id)
                continue


            with open(ssu_hmm_file) as f:
                headers = f.readline().rstrip().split('\t')
                hmm_model_index = headers.index('HMM')
                ssu_length_index = headers.index('SSU gene length')
                contig_length_index = headers.index('Sequence length')

                longest_contig_len = 0
                best_hmm_model = None
                seq_info = None
                seq_id = None
                for line in f:
                    line_split = line.rstrip().split('\t')

                    ssu_query_len = int(line_split[ssu_length_index])
                    if ssu_query_len < min_ssu_length:
                        continue

                    hmm_model = line_split[hmm_model_index]
                    contig_length = int(line_split[contig_length_index])
                    if contig_length < min_ssu_contig:
                        continue

                    if contig_length > longest_contig_len or (best_hmm_model == 'euk' and hmm_model != 'euk'):
                        # preferentially select the longest contig, but aim
                        # for a SSU that had the best match to either the archaeal
                        # or bacterial HMM models as we know these genomes are prokaryotic
                        # (sequences with a best hit to a Eukaryotic model will appear as
                        # Euks when aligned with ssu-align!)
                        longest_contig_len = contig_length
                        best_hmm_model = hmm_model
                        seq_id = line_split[0]
                        seq_info = line.strip()

                if longest_contig_len:
                    ssu_seq_file = os.path.join(genome_dirs[genome_id], 'ssu_gg_2013_08', 'ssu.fna')
                    seqs = seq_io.read_fasta(ssu_seq_file)
                    fout.write('>' + genome_id + ' ' + ';'.join(gtdb_taxonomy[genome_id]) + '\n')
                    fout.write(seqs[seq_id] + '\n')
        fout.close()

        self.logger.info('Identified %d genomes with no 16S sequence meeting filtering criteria.' % no_identified_ssu)

        return ssu_output_file

    def _trim_seqs(self, input_msa, output_msa, remove_identical=False, min_per_taxa=0.7, min_bp=1000):
        """Trim ends of sequences.

        input_msa : str
            File with MSA to trim.
        output_msa : str
            New file with trimmed MSA.
        remove_identical : boolean
            Flag indicating if identical sequence should be removed.
        min_per_taxa : float [0, 1.0]
            Minimum required taxa to retain leading and trailing columns.
        min_bp : int
            Minimum required length to retain sequence.
        """

        # read seqs
        seqs = seq_io.read_fasta(input_msa)

        # filter identical seqs
        identical_seqs = set()
        if remove_identical:
            self.logger.info('Filtering identical sequences.')

            seq_ids = seqs.keys()
            for i in xrange(0, len(seq_ids)):
                seq_id_I = seq_ids[i]

                if seq_id_I in identical_seqs:
                    continue

                for j in xrange(i + 1, len(seqIds)):
                    seq_id_J = seq_ids[j]
                    if seqs[seq_id_I] == seqs[seq_id_J]:
                      self.logger.info('Seq %s and %s are identical.' % (seq_id_I, seq_id_J))
                      identical_seqs.add(seq_id_J)

            self.logger.info('Identified %d of %d sequences as identical.' % (len(identical_seqs), len(seqs)))

        # trim start and end columns to consensus alignment
        first_char = []
        last_char = []
        for seq_id, seq in seqs.iteritems():
            if seq_id in identical_seqs:
                continue

            for i, ch in enumerate(seq):
                if ch != '.' and ch != '-':
                    first_char.append(i)
                    break

            for i in xrange(len(seq) - 1, -1, -1):
                if seq[i] != '.' and seq[i] != '-':
                    last_char.append(i)
                    break

        first_char.sort()
        last_char.sort(reverse=True)

        trim_index = int((len(first_char) * min_per_taxa) + 1)

        start = first_char[trim_index]
        end = last_char[trim_index]

        self.logger.info('Trimming seqs from %d to %d leaving a %dbp length alignment.' % (start, end, end - start + 1))

        short_seq_file = output_msa + '.short'
        fout = open(output_msa, 'w')
        fout_short = open(short_seq_file, 'w')
        num_filtered_seq = 0
        for seq_id, seq in seqs.iteritems():
            if seq_id in identical_seqs:
                continue

            valid_bp = 0
            for i in xrange(start, min(len(seq), end + 1)):
                ch = seq[i]
                if ch != '.' and ch != '-':
                    valid_bp += 1

            if valid_bp >= min_bp:
                fout.write('>' + seq_id + '\n')
                fout.write(seq[start:end + 1] + '\n')
            else:
                self.logger.info('Filtering seq %s with %d of %d (%.1f%%) aligned bases.' % (seq_id, valid_bp, (end - start + 1), valid_bp * 100.0 / (end - start + 1)))
                num_filtered_seq += 1
                fout_short.write('>' + seq_id + '\n')
                fout_short.write(seq[start:end + 1] + '\n')

        fout.close()
        fout_short.close()

        self.logger.info('Filtered %d of %d sequences due to length.' % (num_filtered_seq, len(seqs) - len(identical_seqs)))
        self.logger.info('Short sequence written to: %s' % short_seq_file)

    def _tax_filter(self, ssu_output_file, taxonomy, output_dir):
        """Identify sequence to filter based on taxonomy of best BLAST hit.

        """

        blast = Blast(self.cpus)
        self.logger.info('Creating BLASTN database.')
        blast.create_blastn_db(ssu_output_file)

        self.logger.info('Performing reciprocal BLAST to identify sequences with incongruent taxonomies.')
        blast_table = os.path.join(output_dir, 'blastn.tsv')
        blast.blastn(ssu_output_file,
                     ssu_output_file,
                     blast_table,
                     evalue=1e-10,
                     max_matches=2,
                     output_fmt='custom',
                     task='megablast')

        filter = set()
        order_index = Taxonomy.rank_labels.index('order')
        for hit in blast.read_hit(blast_table, table_fmt='custom'):
            if hit.query_id == hit.subject_id:
                # ignore self hits
                continue

            if hit.perc_identity >= 97 and hit.alignment_len > 800:
                # there is a close hit in the database so verify it has
                # the expected taxonomic order

                order_of_query = taxonomy[hit.query_id][order_index][3:].strip()
                if order_of_query != Taxonomy.rank_prefixes[order_index]:
                    order_of_subject = taxonomy[hit.subject_id][order_index][3:].strip()
                    if order_of_query and order_of_subject and order_of_query != order_of_subject:
                        filter.add(hit.query_id)

        return filter

    def run(self, min_ssu_length,
                    min_ssu_contig,
                    min_quality,
                    max_contigs,
                    min_N50,
                    tax_filter,
                    ncbi_rep_only,
                    user_genomes,
                    genome_list,
                    output_dir):
        """Infer 16S tree spanning select GTDB genomes.

        Parameters
        ----------
        min_ssu_length : int
            Minimum required length of 16S sequences.
        min_ssu_contig : int
            Minimum required length of contig containing 16S sequence.
        min_quality : float [0, 100]
            Minimum genome quality for a genome to be include in tree.
        max_contigs : int
            Maximum number of contigs to include genome.
        min_N50 : int
            Minimum N50 to include genome.
        tax_filter : boolean
            Filter sequences based on incongruent taxonomy classification.
        ncbi_rep_only : boolean
            Restrict tree to NCBI representative and reference genomes.
        user_genomes : boolean
            Include User genomes in tree.
        genome_list : str
            Explict list of genomes to use (ignores --ncbi_rep_only and --user_genomes).
        output_dir : str
            Directory to store results
        """
        
        if genome_list and (ncbi_rep_only or user_genomes):
            self.logger.error("The 'genome_list' flag cannot be used with the 'ncbi_rep_only' and 'user_genomes' flags.")
            sys.exit(-1)

        genome_quality = read_gtdb_metadata(self.gtdb_metadata_file, ['checkm_completeness',
                                                                      'checkm_contamination',
                                                                      'scaffold_count',
                                                                      'n50_contigs'])

        gtdb_taxonomy = read_gtdb_taxonomy(self.gtdb_metadata_file)

        num_user_genomes = 0
        num_ncbi_genomes = 0
        for genome_id in genome_quality:
            if genome_id.startswith('U_'):
                num_user_genomes += 1
            elif genome_id.startswith('RS_') or genome_id.startswith('GB_'):
                num_ncbi_genomes += 1
            else:
                self.logger.warning('Unrecognized genome prefix: %s' % genome_id)
        self.logger.info('Initially considering %d genomes (%d NCBI, %d User).' % (len(genome_quality), num_ncbi_genomes, num_user_genomes))
        
        # filter genomes based on quality and database source
        genomes_in_list = set()
        if genome_list:
            for line in open(genome_list):
                genomes_in_list.add(line.rstrip().split('\t')[0])
            self.logger.info('Restricted genomes to the %d in the genome list.' % len(genomes_in_list))
        else:
            if ncbi_rep_only:
                _accession_to_taxid, _complete_genomes, ncbi_rep_genomes = ncbi.read_refseq_metadata(self.gtdb_metadata_file , keep_db_prefix=True)
                self.logger.info('Considering only NCBI representative or reference genomes.')
                self.logger.info('Identified %d RefSeq genomes.' % len(accession_to_taxid))
                self.logger.info('Identified %d representative or reference genomes.' % len(ncbi_rep_genomes))
                
            if not user_genomes:
                self.logger.info('Filtering User genomes.')

        self.logger.info('Filtering genomes based on specified critieria.')
        new_genomes_to_consider = []
        filtered_genomes = 0
        gq = 0
        sc = 0
        n50 = 0
        for genome_id in genome_quality:                
            if genomes_in_list: 
                if genome_id not in genomes_in_list:
                    continue
            else:
                if not user_genomes and genome_id.startswith('U_'):
                    filtered_genomes += 1
                    continue

                if ncbi_rep_only and genome_id not in ncbi_rep_genomes:
                    filtered_genomes += 1
                    continue
   
            comp, cont, scaffold_count, n50_contigs = genome_quality.get(genome_id, [-1, -1, -1])
            q = float(comp) - float(cont)
            if q < min_quality or int(scaffold_count) > max_contigs or int(n50_contigs) < min_N50:
                if q < min_quality:
                    gq += 1
                    
                if int(scaffold_count) > max_contigs:
                    sc += 1
                    
                if int(n50_contigs) < min_N50:
                    n50 += 1
            
                filtered_genomes += 1
                continue
                
            new_genomes_to_consider.append(genome_id)

        genomes_to_consider = new_genomes_to_consider
        self.logger.info('Filtered %d genomes (%d on genome quality, %d on number of contigs, %d on N50).' % (filtered_genomes, gq, sc, n50))
        self.logger.info('Considering %d genomes after filtering.' % len(genomes_to_consider))

        # get SSU sequences for genomes
        ssu_output_file = self._get_ssu_seqs(min_ssu_length,
                                             min_ssu_contig,
                                             gtdb_taxonomy,
                                             genomes_to_consider,
                                             output_dir)

        # identify erroneous SSU sequences
        if tax_filter:
            self.logger.info('Filtering sequences with incongruent taxonomy strings.')
            filter = self._tax_filter(ssu_output_file, gtdb_taxonomy, output_dir)

            self.logger.info('Filtered %d sequences.' % len(filter))
            print filter

            if len(filter) > 0:
                ssu_filtered_output = os.path.join(output_dir, 'gtdb_ssu.filtered.fna')
                fout = open(ssu_filtered_output, 'w')
                for seq_id, seq, annotation in seq_io.read_seq(ssu_output_file, keep_annotation=True):
                    if seq_id not in filter:
                        fout.write('>' + seq_id + ' ' + annotation + '\n')
                        fout.write(seq + '\n')
                fout.close()

                ssu_output_file = ssu_filtered_output

        # align sequences
        ssu_align_dir = os.path.join(output_dir, 'ssu_align')
        os.system('ssu-align --dna %s %s' % (ssu_output_file, ssu_align_dir))
        os.system('ssu-mask --afa %s' % ssu_align_dir)

        # trim sequences
        for domain in ['archaea', 'bacteria']:
            input_msa = os.path.join(ssu_align_dir, 'ssu_align.' + domain + '.mask.afa')
            trimmed_msa = os.path.join(output_dir, domain + '.trimmed.fna')
            self._trim_seqs(input_msa, trimmed_msa)

            # infer tree
            output_tree = os.path.join(output_dir, domain + '.tree')
            os.system('FastTreeMP -nosupport -nt -gamma %s > %s' % (trimmed_msa, output_tree))



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging

import dendropy

from biolib.common import check_file_exists, make_sure_path_exists, is_float
from biolib.external.execute import check_dependencies
from biolib.taxonomy import Taxonomy

from genometreetk.exceptions import GenomeTreeTkError
from genometreetk.trusted_genome_workflow import TrustedGenomeWorkflow
from genometreetk.dereplication_workflow import DereplicationWorkflow
from genometreetk.marker_workflow import MarkerWorkflow
from genometreetk.infer_workflow import InferWorkflow
from genometreetk.ssu_workflow import SSU_Workflow
from genometreetk.bootstrap import Bootstrap
from genometreetk.jackknife_markers import JackknifeMarkers
from genometreetk.jackknife_taxa import JackknifeTaxa
from genometreetk.combine_support import CombineSupport
from genometreetk.reroot_tree import RerootTree
from genometreetk.representatives import Representatives
from genometreetk.cluster import Cluster
from genometreetk.common import read_gtdb_metadata


class OptionsParser():
    def __init__(self):
        """Initialization"""
        self.logger = logging.getLogger()

    def _read_config_file(self):
        """Read configuration info."""

        cfg_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'genometreetk.cfg')

        d = {}
        for line in open(cfg_file):
            key, value = line.split('=')
            d[key] = value.strip()

        return d

    def trusted(self, options):
        """Determine trusted genomes."""

        check_file_exists(options.metadata_file)

        trusted_genome_workflow = TrustedGenomeWorkflow()

        trusted_genome_workflow.run(options.metadata_file,
                                        options.trusted_comp,
                                        options.trusted_cont,
                                        options.max_contigs,
                                        options.min_N50,
                                        options.refseq_rep,
                                        options.trusted_genomes_file)

        self.logger.info('Trusted genome list written to: %s' % options.trusted_genomes_file)

    def dereplicate(self, options):
        """Dereplicate genomes based on taxonomy."""

        if options.trusted_genomes_file:
            check_file_exists(options.trusted_genomes_file)

        try:
            dereplication_workflow = DereplicationWorkflow()

            dereplication_workflow.run(options.max_species,
                                       options.trusted_genomes_file,
                                       None,
                                       None,
                                       options.derep_genome_file)
        except GenomeTreeTkError as e:
            print e.message
            raise SystemExit

        self.logger.info('Dereplicated genome list written to: %s' % options.derep_genome_file)

    def markers(self, options):
        """Determine marker genes."""

        make_sure_path_exists(options.output_dir)

        config_data = self._read_config_file()

        self.logger.error('NEED TO FIX ISSUE WITH GENOME DIR FILES!')
        sys.exit(-1)

        marker_workflow = MarkerWorkflow(config_data['genome_dir_file'],
                                            config_data['pfam_model_file'],
                                            config_data['tigrfams_model_dir'],
                                            options.cpus)

        phylo_hmm_out = marker_workflow.run(options.ingroup_file,
                                                options.ubiquity,
                                                options.single_copy,
                                                options.redundancy,
                                                options.min_support,
                                                options.min_per_taxa,
                                                options.perc_markers,
                                                options.restict_marker_list,
                                                options.output_dir)

        self.logger.info('Marker genes written to: %s' % phylo_hmm_out)

    def infer(self, options):
        """Infer genome tree."""

        check_file_exists(options.genome_id_file)
        check_file_exists(options.marker_id_file)
        make_sure_path_exists(options.output_dir)

        self.logger.error('NEED TO FIX ISSUE WITH GENOME DIR FILES!')
        sys.exit(-1)

        config_data = self._read_config_file()
        infer_workflow = InferWorkflow(config_data['genome_dir_file'],
                                            config_data['pfam_model_file'],
                                            config_data['tigrfams_model_dir'],
                                            options.cpus)

        infer_workflow.run(options.genome_id_file,
                                options.marker_id_file,
                                options.model,
                                options.output_dir)

        self.logger.info('Results written to: %s' % options.output_dir)

    def ssu_tree(self, options):
        """Infer 16S tree spanning GTDB genomes."""

        check_dependencies(['ssu-align', 'ssu-mask', 'FastTreeMP', 'blastn'])

        check_file_exists(options.gtdb_metadata_file)
        check_file_exists(options.gtdb_dir_file)
        make_sure_path_exists(options.output_dir)

        ssu_workflow = SSU_Workflow(options.gtdb_metadata_file,
                                    options.gtdb_dir_file,
                                    options.cpus)
        ssu_workflow.run(options.min_ssu_length,
                         options.min_ssu_contig,
                         options.min_quality,
                         options.max_contigs,
                         options.min_N50,
                         not options.disable_tax_filter,
                         options.ncbi_rep_only,
                         options.user_genomes,
                         options.genome_list,
                         options.output_dir)

        self.logger.info('Results written to: %s' % options.output_dir)

    def bootstrap(self, options):
        """Bootstrap multiple sequence alignment."""

        check_file_exists(options.input_tree)
        check_file_exists(options.msa_file)
        make_sure_path_exists(options.output_dir)

        bootstrap = Bootstrap(options.cpus)
        output_tree = bootstrap.run(options.input_tree,
                                    options.msa_file,
                                    options.num_replicates,
                                    options.model,
                                    options.base_type,
                                    options.fraction,
                                    options.output_dir)

        self.logger.info('Bootstrapped tree written to: %s' % output_tree)

    def jk_markers(self, options):
        """Jackknife marker genes."""

        check_file_exists(options.input_tree)
        check_file_exists(options.msa_file)
        make_sure_path_exists(options.output_dir)

        jackknife_markers = JackknifeMarkers(options.cpus)
        output_tree = jackknife_markers.run(options.input_tree,
                                                options.msa_file,
                                                options.gene_length_file,
                                                options.perc_markers,
                                                options.num_replicates,
                                                options.model,
                                                options.output_dir)

        self.logger.info('Jackknifed marker tree written to: %s' % output_tree)

    def jk_taxa(self, options):
        """Jackknife taxa."""

        check_file_exists(options.input_tree)
        check_file_exists(options.msa_file)
        make_sure_path_exists(options.output_dir)

        jackknife_taxa = JackknifeTaxa(options.cpus)
        output_tree = jackknife_taxa.run(options.input_tree,
                                            options.msa_file,
                                            options.outgroup_ids,
                                            options.perc_taxa,
                                            options.num_replicates,
                                            options.model,
                                            options.output_dir)

        self.logger.info('Jackknifed taxa tree written to: %s' % output_tree)

    def combine(self, options):
        """Combine support values into a single tree."""

        combineSupport = CombineSupport()
        combineSupport.run(options.support_type,
                            options.bootstrap_tree,
                            options.jk_marker_tree,
                            options.jk_taxa_tree,
                            options.output_tree)

    def support_wf(self, options):
        """"Perform entire tree support workflow."""

        self.bootstrap(options)
        self.jk_markers(options)
        self.jk_taxa(options)
        self.combine(options)

    def midpoint(self, options):
        """"Midpoint root tree."""

        reroot = RerootTree()
        reroot.midpoint(options.input_tree, options.output_tree)

    def outgroup(self, options):
        """Reroot tree with outgroup."""

        check_file_exists(options.taxonomy_file)

        self.logger.info('Identifying genomes from the specified outgroup.')
        outgroup = set()
        for genome_id, taxa in Taxonomy().read(options.taxonomy_file).iteritems():
            if options.outgroup_taxon in taxa:
                outgroup.add(genome_id)
        self.logger.info('Identifying %d genomes in the outgroup.' % len(outgroup))

        reroot = RerootTree()
        reroot.root_with_outgroup(options.input_tree, options.output_tree, outgroup)

    def refseq_representatives(self, options):
        """Determine representative genomes in RefSeq."""

        check_file_exists(options.metadata_file)

        try:
            rep_workflow = DereplicationWorkflow()

            rep_workflow.run(options.max_species,
                                       None,
                                       options.metadata_file,
                                       options.min_rep_comp,
                                       options.max_rep_cont,
                                       options.min_quality,
                                       options.max_contigs,
                                       options.min_N50,
                                       options.max_ambiguous,
                                       options.strict_filtering,
                                       options.rep_genome_file)
        except GenomeTreeTkError as e:
            print e.message
            raise SystemExit

        self.logger.info('RefSeq representative genomes written to: %s' % options.rep_genome_file)

    def representatives(self, options):
        """Determine additional representatives genomes."""

        check_file_exists(options.ar_msa_file)
        check_file_exists(options.bac_msa_file)
        check_file_exists(options.refseq_representatives)
        check_file_exists(options.metadata_file)

        try:
            representatives = Representatives()
            representatives.run(options.refseq_representatives,
                                    options.trusted_user_file,
                                    options.ar_msa_file,
                                    options.bac_msa_file,
                                    options.threshold,
                                    options.min_rep_comp,
                                    options.max_rep_cont,
                                    options.min_quality,
                                    options.max_contigs,
                                    options.min_N50,
                                    options.max_ambiguous,
                                    options.metadata_file,
                                    options.rep_genome_file)

            self.logger.info('Representative genomes written to: %s' % options.rep_genome_file)

        except GenomeTreeTkError as e:
            print e.message
            raise SystemExit

    def aai_cluster(self, options):
        """Cluster genomes based on AAI."""

        check_file_exists(options.ar_msa_file)
        check_file_exists(options.bac_msa_file)
        check_file_exists(options.representatives)
        check_file_exists(options.metadata_file)

        try:
            cluster = Cluster(options.cpus)
            cluster.run(options.representatives,
                        options.trusted_user_file,
                        options.ar_msa_file,
                        options.bac_msa_file,
                        options.threshold,
                        options.metadata_file,
                        options.cluster_file)

            self.logger.info('Clustering information written to: %s' % options.cluster_file)

        except GenomeTreeTkError as e:
            print e.message
            raise SystemExit

    def validate(self, options):
        """Check taxonomy file is formatted as expected."""

        check_file_exists(options.input_taxonomy)

        taxonomy = Taxonomy()
        t = taxonomy.read(options.input_taxonomy)

        taxonomy.validate(t,
                          check_prefixes=True,
                          check_ranks=True,
                          check_hierarchy=True,
                          check_species=True,
                          report_errors=True)

        self.logger.info('Finished performing validation tests.')

    def fill_ranks(self, options):
        """Ensure taxonomy strings contain all 7 canonical ranks."""

        check_file_exists(options.input_taxonomy)

        fout = open(options.output_taxonomy, 'w')
        taxonomy = Taxonomy()
        t = taxonomy.read(options.input_taxonomy)

        for genome_id, taxon_list in t.iteritems():
            full_taxon_list = taxonomy.fill_missing_ranks(taxon_list)

            taxonomy_str = ';'.join(full_taxon_list)
            if not taxonomy.check_full(taxonomy_str):
                sys.exit(-1)

            fout.write('%s\t%s\n' % (genome_id, taxonomy_str))

        fout.close()

        self.logger.info('Revised taxonomy written to: %s' % options.output_taxonomy)

    def binomial(self, options):
        """Ensure species are designated using binomial nomenclature."""

        check_file_exists(options.input_taxonomy)

        fout = open(options.output_taxonomy, 'w')
        taxonomy = Taxonomy()
        t = taxonomy.read(options.input_taxonomy)

        for genome_id, taxon_list in t.iteritems():
            taxonomy_str = ';'.join(taxon_list)
            if not taxonomy.check_full(taxonomy_str):
                sys.exit(-1)

            genus = taxon_list[5][3:]
            species = taxon_list[6][3:]
            if species and genus not in species:
                taxon_list[6] = 's__' + genus + ' ' + species
                taxonomy_str = ';'.join(taxon_list)

            fout.write('%s\t%s\n' % (genome_id, taxonomy_str))

        fout.close()


        self.logger.info('Revised taxonomy written to: %s' % options.output_taxonomy)

    def propagate(self, options):
        """Propagate labels to all genomes in a cluster."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.metadata_file)

        # get representative genome information
        rep_metadata = read_gtdb_metadata(options.metadata_file, ['gtdb_representative',
                                                                  'gtdb_clustered_genomes'])

        taxonomy = Taxonomy()
        explict_tax = taxonomy.read(options.input_taxonomy)
        expanded_taxonomy = {}
        incongruent_count = 0
        for genome_id, taxon_list in explict_tax.iteritems():
            taxonomy_str = ';'.join(taxon_list)

            # Propagate taxonomy strings if genome is a representatives. Also, determine
            # if genomes clustered together have compatible taxonomies. Note that a genome
            # may not have metadata as it is possible a User  has removed a genome that is
            # in the provided taxonomy file.
            _rep_genome, clustered_genomes = rep_metadata.get(genome_id, (None, None))
            if clustered_genomes:  # genome is a representative
                clustered_genome_ids = clustered_genomes.split(';')

                # get taxonomy of all genomes in cluster with a specified taxonomy
                clustered_genome_tax = {}
                for cluster_genome_id in clustered_genome_ids:
                    if cluster_genome_id == genome_id:
                        continue

                    if cluster_genome_id not in rep_metadata:
                        continue  # genome is no longer in the GTDB so ignore it

                    if cluster_genome_id in explict_tax:
                        clustered_genome_tax[cluster_genome_id] = explict_tax[cluster_genome_id]

                # determine if representative and clustered genome taxonomy strings are congruent
                working_cluster_taxonomy = list(taxon_list)
                incongruent_with_rep = False
                for cluster_genome_id, cluster_tax in clustered_genome_tax.iteritems():
                    if incongruent_with_rep:
                        working_cluster_taxonomy = list(taxon_list)  # default to rep taxonomy
                        break

                    for r in xrange(0, len(Taxonomy.rank_prefixes)):
                        if cluster_tax[r] == Taxonomy.rank_prefixes[r]:
                            break  # no more taxonomy information to consider

                        if cluster_tax[r] != taxon_list[r]:
                            if taxon_list[r] == Taxonomy.rank_prefixes[r]:
                                # clustered genome has a more specific taxonomy string which
                                # should be propagate to the representative if all clustered
                                # genomes are in agreement
                                if working_cluster_taxonomy[r] == Taxonomy.rank_prefixes[r]:
                                    # make taxonomy more specific based on genomes in cluster
                                    working_cluster_taxonomy[r] = cluster_tax[r]
                                elif working_cluster_taxonomy[r] != cluster_tax[r]:
                                    # not all genomes agree on the assignment of this rank so leave it unspecified
                                    working_cluster_taxonomy[r] = Taxonomy.rank_prefixes[r]
                                    break
                            else:
                                # genomes in cluster have incongruent taxonomies so defer to representative
                                self.logger.warning("Genomes in cluster have incongruent taxonomies.")
                                self.logger.warning("Representative %s: %s" % (genome_id, taxonomy_str))
                                self.logger.warning("Clustered genome %s: %s" % (cluster_genome_id, ';'.join(cluster_tax)))
                                self.logger.warning("Deferring to taxonomy specified for representative.")

                                incongruent_count += 1
                                incongruent_with_rep = True
                                break

                cluster_taxonomy_str = ';'.join(working_cluster_taxonomy)

                # assign taxonomy to representative and all genomes in the cluster
                expanded_taxonomy[genome_id] = cluster_taxonomy_str
                for cluster_genome_id in clustered_genome_ids:
                    expanded_taxonomy[cluster_genome_id] = cluster_taxonomy_str
            else:
                if genome_id in expanded_taxonomy:
                    # genome has already been assigned a taxonomy based on its representative
                    pass
                else:
                    # genome is a singleton
                    expanded_taxonomy[genome_id] = taxonomy_str


        self.logger.info('Identified %d clusters with incongruent taxonomies.' % incongruent_count)

        fout = open(options.output_taxonomy, 'w')
        for genome_id, taxonomy_str in expanded_taxonomy.iteritems():
            fout.write('%s\t%s\n' % (genome_id, taxonomy_str))
        fout.close()

        self.logger.info('Taxonomy written to: %s' % options.output_taxonomy)

    def strip(self, options):
        """Remove taxonomic labels from tree."""

        check_file_exists(options.input_tree)

        outgroup_in_tree = set()
        tree = dendropy.Tree.get_from_path(options.input_tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)

        for node in tree.internal_nodes():
            if node.label:
                if ':' in node.label:
                    support, _taxa = node.label.split(':')
                    node.label = support

        tree.write_to_path(options.output_tree,
                            schema='newick',
                            suppress_rooting=True,
                            unquoted_underscores=True)

        self.logger.info('Stripped tree written to: %s' % options.output_tree)
        
    def pull(self, options):
        """Create taxonomy file from a decorated tree."""

        check_file_exists(options.input_tree)

        taxonomy = Taxonomy().read_from_tree(options.input_tree)
        Taxonomy().write(taxonomy, options.output_taxonomy)
            
        self.logger.info('Stripped tree written to: %s' % options.output_taxonomy)

    def diff(self, options):
        """Compare two taxonomy files."""

        check_file_exists(options.input_taxonomy1)
        check_file_exists(options.input_taxonomy2)

        taxonomy1 = Taxonomy().read(options.input_taxonomy1)
        taxonomy2 = Taxonomy().read(options.input_taxonomy2)

        all_taxon_ids = set(taxonomy1.keys()).union(taxonomy2.keys())

        rank_index = Taxonomy.rank_labels.index(options.rank)
        for taxon_id in all_taxon_ids:
            if options.report_missing_taxa:
                if taxon_id not in taxonomy1:
                    print 'Missing in taxonomy 1: %s' % taxon_id
                elif taxon_id not in taxonomy2:
                    print 'Missing in taxonomy 2: %s' % taxon_id

            if taxon_id in taxonomy1 and taxon_id in taxonomy2:
                taxon1 = taxonomy1[taxon_id][rank_index]
                taxon2 = taxonomy2[taxon_id][rank_index]

                if taxon1 != taxon2:
                    if options.report_missing_ranks or (taxon1[3:] and taxon2[3:]):
                        print 'Different taxon for %s: %s %s' % (taxon_id, taxon1, taxon2)

        print 'Done.'

    def phylogenetic_diversity(self, options):
        """Calculate phylogenetic diversity of named groups."""

        check_file_exists(options.decorated_tree)

        tree = dendropy.Tree.get_from_path(options.decorated_tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)

        total_pd = 0
        pds = {}
        for node in tree.preorder_node_iter():
            if node.parent_node is not None:
                total_pd += node.edge.length

            if not node.label:
                continue

            taxon = None
            if ':' in node.label:
                _support, taxon = node.label.split(':')
            else:
                if not is_float(node.label):
                    taxon = node.label

            if taxon:
                taxon_pd = 0
                for nn in node.postorder_iter():
                    if nn != node:
                        taxon_pd += nn.edge.length

                print '%s\t%f' % (taxon, taxon_pd)

        print '%s\t%f' % ('Total branch length', total_pd)

    def parse_options(self, options):
        """Parse user options and call the correct pipeline(s)"""

        logging.basicConfig(format='', level=logging.INFO)

        check_dependencies(('FastTree', 'hmmsearch'))

        if options.subparser_name == 'trusted':
            self.trusted(options)
        elif options.subparser_name == 'dereplicate':
            self.dereplicate(options)
        elif options.subparser_name == 'markers':
            self.markers(options)
        elif options.subparser_name == 'infer':
            self.infer(options)
        elif options.subparser_name == 'ssu_tree':
            self.ssu_tree(options)
        elif options.subparser_name == 'bootstrap':
            self.bootstrap(options)
        elif options.subparser_name == 'jk_markers':
            self.jk_markers(options)
        elif options.subparser_name == 'jk_taxa':
            self.jk_taxa(options)
        elif options.subparser_name == 'combine':
            self.combine(options)
        elif options.subparser_name == 'midpoint':
            self.midpoint(options)
        elif options.subparser_name == 'outgroup':
            self.outgroup(options)
        elif options.subparser_name == 'refseq_reps':
            self.refseq_representatives(options)
        elif options.subparser_name == 'reps':
            self.representatives(options)
        elif options.subparser_name == 'aai_cluster':
            self.aai_cluster(options)
        elif options.subparser_name == 'validate':
            self.validate(options)
        elif options.subparser_name == 'binomial':
            self.binomial(options)
        elif options.subparser_name == 'binomial':
            self.binomial(options)
        elif options.subparser_name == 'propagate':
            self.propagate(options)
        elif options.subparser_name == 'fill_ranks':
            self.fill_ranks(options)
        elif options.subparser_name == 'strip':
            self.strip(options)
        elif options.subparser_name == 'pull':
            self.pull(options)
        elif options.subparser_name == 'diff':
            self.diff(options)
        elif options.subparser_name == 'pd':
            self.phylogenetic_diversity(options)
        else:
            self.logger.error('  [Error] Unknown GenomeTreeTk command: ' + options.subparser_name + '\n')
            sys.exit()

        return 0



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import sys
import logging
import random
import operator
from collections import defaultdict

from biolib.taxonomy import Taxonomy

from genometreetk.common import (read_gtdb_metadata,
                                 read_gtdb_ncbi_taxonomy,
                                 read_gtdb_ncbi_organism_name,
                                 read_gtdb_taxonomy,
                                 read_gtdb_ncbi_type_strain,
                                 species_label)
import genometreetk.ncbi as ncbi


class DereplicationWorkflow(object):
    """Dereplicate genomes based on taxonomy.

    Each named species is dereplicated to a fixed number of
    taxa, taking care to retain all genomes marked as a
    'reference' or 'representative' at NCBI. Preference
    is then given to genomes marked as type strains at
    NCBI.

    All taxa without a fully qualified species name are
    retained.
    """

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger()

    def _dereplicate(self, assemble_accessions,
                            max_species,
                            species_labels,
                            representative_genomes,
                            complete_genomes,
                            ncbi_type_strains,
                            lpsn_type_strains,
                            trusted_accessions,
                            genome_quality):
        """Dereplicate genomes based on taxonomy.

        Parameters
        ----------
        assemble_accessions : set
            All assemble accessions to consider.
        max_species : int
            Maximum number of genomes of the same species to retain.
        species_labels : d[assembly_accession] -> species
            Species label for each genome.
        representative_genomes : set
            Set of representative genomes to prioritize.
        complete_genomes : set
            Set of complete genomes to prioritize.
        ncbi_type_strains : set
            Set of genomes marked as type strains at NCBI.
        lpsn_type_strains : d[ncbi_species] -> set of genome IDs
            Genomes marked as type strains at LPSN.
        trusted_accessions : set
            Set of trusted genomes.
        genome_quality : d[genome_id] -> comp - cont
            Quality of each genome.

        Returns
        -------
        set
            Dereplicate set of assemblies.
        """

        # determine genomes belonging to each named species
        retained_reps = 0
        species = defaultdict(set)
        genomes_to_retain = set()
        for genome_id in assemble_accessions:
            sp = species_labels.get(genome_id, None)

            if not trusted_accessions or genome_id in trusted_accessions:
                if not sp:
                    if genome_id in representative_genomes:
                        retained_reps += 1
                    genomes_to_retain.add(genome_id)
                else:
                    species[sp].add(genome_id)

        self.logger.info('Retained %d genomes without a species designation.' % (len(genomes_to_retain) - retained_reps))

        # dereplicate species
        additional_reps = 0
        for sp, genome_ids in species.iteritems():
            representatives = genome_ids.intersection(representative_genomes)

            if len(genome_ids) > max_species:
                selected_genomes = set()

                # get different classes of genomes
                complete = genome_ids.intersection(complete_genomes)
                type_strains = genome_ids.intersection(ncbi_type_strains)
                complete_type_strains = type_strains.intersection(complete)
                complete = complete - complete_type_strains

                # select all representative genomes
                if representatives:
                    selected_genomes.update(representatives)
                    retained_reps += len(representatives)
                    genome_ids.difference_update(representatives)
                    type_strains.difference_update(representatives)
                    complete_type_strains.difference_update(representatives)
                    complete.difference_update(representatives)

                # select genomes based on estimated genome quality
                complete_quality = {k:genome_quality[k] for k in complete}
                complete_quality_sorted = sorted(complete_quality.items(), key=operator.itemgetter(1), reverse=True)

                type_strains_quality = {k:genome_quality[k] for k in type_strains}
                type_strains_quality_sorted = sorted(type_strains_quality.items(), key=operator.itemgetter(1), reverse=True)

                complete_type_strains_quality = {k:genome_quality[k] for k in complete_type_strains}
                complete_type_strains_quality_sorted = sorted(complete_type_strains_quality.items(), key=operator.itemgetter(1), reverse=True)

                # try to select a complete type strain, otherwise just take
                # any type strain if one exists
                if len(selected_genomes) < max_species:
                    if complete_type_strains:
                        selected_type_strain = [x[0] for x in complete_type_strains_quality_sorted[0:1]] # random.sample(complete_type_strains, 1)
                        selected_genomes.update(selected_type_strain)
                    elif type_strains:
                        selected_type_strain = [x[0] for x in type_strains_quality_sorted[0:1]]# random.sample(type_strains, 1)
                        selected_genomes.update(selected_type_strain)

                # grab as many complete genomes as possible
                if len(selected_genomes) < max_species and complete:
                    genomes_to_select = min(len(complete), max_species - len(selected_genomes))
                    selected_complete_genomes = [x[0] for x in complete_quality_sorted[0:genomes_to_select]] #random.sample(complete, genomes_to_select)
                    selected_genomes.update(selected_complete_genomes)

                # grab incomplete genomes to get to the desired number of genomes
                if len(selected_genomes) < max_species and genome_ids:
                    genome_ids.difference_update(selected_genomes)
                    genome_ids_quality = {k:genome_quality[k] for k in genome_ids}
                    genome_ids_quality_sorted = sorted(genome_ids_quality.items(), key=operator.itemgetter(1), reverse=True)
                    
                    genomes_to_select = min(len(genome_ids), max_species - len(selected_genomes))
                    rnd_additional_genomes = [x[0] for x in genome_ids_quality_sorted[0:genomes_to_select]] #random.sample(genome_ids, genomes_to_select)
                    selected_genomes.update(rnd_additional_genomes)

                genomes_to_retain.update(selected_genomes)
                additional_reps += len(selected_genomes) - len(representatives)
            else:
                genomes_to_retain.update(genome_ids)
                retained_reps += len(representatives)
                additional_reps += len(genome_ids) - len(representatives)
                
        self.logger.info('Retained %d RefSeq representatives.' % retained_reps)
        self.logger.info('Retained %d additional genomes in order to establish %s genomes per species.' % (additional_reps, max_species))
      
        # make sure to select at least one LPSN type strain for each species
        lpsn_genomes = 0
        for sp, genome_ids in lpsn_type_strains.iteritems():
            if len(genome_ids.intersection(genomes_to_retain)) >= 1:
                # an LPSN type strain has already been selected for this species
                continue
                
            # select genome with the highest quality
            genome_ids_quality = {k:genome_quality[k] for k in genome_ids}
            genome_ids_quality_sorted = sorted(genome_ids_quality.items(), key=operator.itemgetter(1), reverse=True)
            genomes_to_retain.add(genome_ids_quality_sorted[0][0])
            lpsn_genomes += 1

        self.logger.info('Retained %d additional genomes to ensure a LPSN type strain in each species.' % lpsn_genomes)

        return genomes_to_retain

    def run(self, max_species,
                    trusted_genomes_file,
                    metadata_file,
                    min_rep_comp,
                    max_rep_cont,
                    min_quality,
                    max_contigs,
                    min_N50,
                    max_ambiguous,
                    strict_filtering,
                    output_file):
        """Dereplicate genomes to a specific number per named species.

        Parameters
        ----------
        max_species : int
            Maximum number of genomes of the same species to retain.
        trusted_genomes_file:
            File containing list of trusted genomes.
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        min_rep_comp : float [0, 100]
            Minimum completeness for a genome to be a representative.
        max_rep_cont : float [0, 100]
            Maximum contamination for a genome to be a representative.
        min_quality : float [0, 100]
            Minimum genome quality (comp-4*cont) for a genome to be a representative.
        max_contigs : int
            Maximum number of contigs for a genome to be a representative.
        min_N50 : int
            Minimum N50 for a genome to be a representative.
        max_ambiguous : int
            Maximum number of ambiguous bases for a genome to be a representative.
        strict_filtering : boolean
            If True apply filtering to all genomes, otherise apply lenient 
            filtering to genomes where the chromosome and plasmids are reported 
            as complete.
        output_file : str
            Output file to contain list of dereplicated genomes.
        """

        accession_to_taxid, complete_genomes, representative_genomes = ncbi.read_refseq_metadata(metadata_file, keep_db_prefix=True)
        self.logger.info('Identified %d RefSeq genomes.' % len(accession_to_taxid))
        self.logger.info('Identified %d representative or reference genomes.' % len(representative_genomes))
        self.logger.info('Identified %d complete genomes.' % len(complete_genomes))

        gtdb_taxonomy = read_gtdb_taxonomy(metadata_file)
        ncbi_taxonomy = read_gtdb_ncbi_taxonomy(metadata_file)
        ncbi_organism_names = read_gtdb_ncbi_organism_name(metadata_file)
        species = species_label(gtdb_taxonomy, ncbi_taxonomy, ncbi_organism_names)

        self.logger.info('Identified %d genomes with a GTDB or NCBI species names.' % len(species))

        trusted_accessions = set()
        if trusted_genomes_file:
            trusted_accessions = self._read_trusted_genomes(trusted_genomes_file)

        # get genome quality
        genomes_to_consider = accession_to_taxid.keys()
        filtered_reps = 0

        genome_stats = read_gtdb_metadata(metadata_file, ['checkm_completeness',
                                                            'checkm_contamination',
                                                            'contig_count',
                                                            'n50_contigs',
                                                            'ambiguous_bases',
                                                            'scaffold_count',
                                                            'ssu_count',
                                                            'ncbi_molecule_count',
                                                            'ncbi_unspanned_gaps',
                                                            'ncbi_genome_representation',
                                                            'ncbi_spanned_gaps',
                                                            'ncbi_assembly_level',
                                                            'ncbi_taxonomy',
                                                            'lpsn_strain'])
        missing_quality = set(accession_to_taxid.keys()) - set(genome_stats.keys())
        if missing_quality:
            self.logger.error('There are %d genomes without metadata information.' % len(missing_quality))
            self.exit(-1)

        lpsn_type_strains = defaultdict(set)
        new_genomes_to_consider = []
        genome_quality = {}
        for genome_id in accession_to_taxid.keys():
            stats = genome_stats[genome_id]
            comp = stats.checkm_completeness
            cont = stats.checkm_contamination
            
            if (comp >= min_rep_comp
                    and cont <= max_rep_cont
                    and (comp - 4*cont) >= min_quality
                    and stats.contig_count <= max_contigs
                    and stats.n50_contigs >= min_N50
                    and stats.ambiguous_bases <= max_ambiguous):
                        new_genomes_to_consider.append(genome_id)
                        genome_quality[genome_id] = comp - cont
                        if stats.lpsn_strain:
                            ncbi_species = stats.ncbi_taxonomy.split(';')[6].strip()
                            lpsn_type_strains[ncbi_species].add(genome_id)
            elif not strict_filtering:
                # check if genome appears to consist of only an unspanned
                # chromosome and unspanned plasmids and thus can be 
                # subjected to a more lenient quality check
                if (stats.ncbi_assembly_level in ['Complete Genome', 'Chromosome']
                    and stats.ncbi_genome_representation == 'full'
                    and stats.scaffold_count == stats.ncbi_molecule_count
                    and stats.ncbi_unspanned_gaps == 0
                    and stats.ncbi_spanned_gaps <= 10
                    and stats.ambiguous_bases <= 100
                    and stats.ssu_count >= 1):
                    
                    # apply lenient quality check 
                    if comp >= 50 and cont <= 15:
                        new_genomes_to_consider.append(genome_id)
                        genome_quality[genome_id] = comp - cont
                        if stats.lpsn_strain:
                            ncbi_species = stats.ncbi_taxonomy.split(';')[6].strip()
                            lpsn_type_strains[ncbi_species].add(genome_id)
            
            # check if a representative at NCBI is being filtered
            if genome_id in representative_genomes and genome_id not in new_genomes_to_consider:
                self.logger.warning('Filtered RefSeq representative %s with comp=%.2f, cont=%.2f, contigs=%d, N50=%d' % (genome_id, 
                                                                                                                            comp, 
                                                                                                                            cont, 
                                                                                                                            stats.contig_count, 
                                                                                                                            stats.n50_contigs))
                filtered_reps += 1

        genomes_to_consider = new_genomes_to_consider
        self.logger.info('Filtered %d representative or reference genome based on genome quality.' % filtered_reps)
        self.logger.info('Considering %d genomes after filtering for genome quality.' % (len(genomes_to_consider)))

        ncbi_type_strains = read_gtdb_ncbi_type_strain(metadata_file)
        self.logger.info('Identified %d genomes marked as type strains at NCBI.' % len(ncbi_type_strains))
        self.logger.info('Identified %d genomes marked as type strains at LPSN.' % sum([len(x) for x in lpsn_type_strains.values()]))

        genomes_to_retain = self._dereplicate(genomes_to_consider,
                                                max_species,
                                                species,
                                                representative_genomes,
                                                complete_genomes,
                                                ncbi_type_strains,
                                                lpsn_type_strains,
                                                trusted_accessions,
                                                genome_quality)

        self.logger.info('Retained %d genomes.' % len(genomes_to_retain))

        if not trusted_genomes_file:
            trusted_genomes_file = ''

        fout = open(output_file, 'w')
        fout.write('# Selection criteria:\n')
        fout.write('# Maximum species: %d\n' % max_species)
        fout.write('# Trusted genomes file: %s\n' % trusted_genomes_file)
        fout.write('# Genome quality metadata file: %s\n' % str(metadata_file))
        fout.write('# Min. representative completeness: %s\n' % str(min_rep_comp))
        fout.write('# Max. representative contamination: %s\n' % str(max_rep_cont))
        fout.write('#\n')
        fout.write('# Genome Id\tGTDB Taxonomy\tNCBI Taxonomy\tType strain\tComplete\tRepresentative\n')
        for assembly_accession in genomes_to_retain:
            representative = 'yes' if assembly_accession in representative_genomes else 'no'
            complete = 'yes' if assembly_accession in complete_genomes else 'no'
            ts = 'yes' if assembly_accession in ncbi_type_strains else 'no'
            gtdb_taxa_str = ';'.join(gtdb_taxonomy.get(assembly_accession, Taxonomy.rank_prefixes))
            ncbi_taxa_str = ';'.join(ncbi_taxonomy.get(assembly_accession, Taxonomy.rank_prefixes))

            if assembly_accession.startswith('GCF_'):
                assembly_accession = 'RS_' + assembly_accession
            elif assembly_accession.startswith('GCA_'):
                assembly_accession = 'GB_' + assembly_accession

            fout.write('%s\t%s\t%s\t%s\t%s\t%s\n' % (assembly_accession,
                                                         gtdb_taxa_str,
                                                         ncbi_taxa_str,
                                                         ts,
                                                         complete,
                                                         representative))
        fout.close()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging

from genometreetk.common import read_gtdb_taxonomy, read_gtdb_ncbi_taxonomy
import genometreetk.ncbi as ncbi

import csv


class TrustedGenomeWorkflow(object):
    """Determine trusted genomes based on genome statistics."""

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger()

    def _trusted_genomes(self, metadata_file,
                                trusted_comp,
                                trusted_cont,
                                max_contigs,
                                min_N50):
        """Identify trusted genomes.

        Parameters
        ----------
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        trusted_comp : float [0, 100]
            Minimum completeness of trusted genomes.
        trusted_cont : float [0, 100]
            Maximum contamination of trusted genomes.
        max_contigs : int
            Maximum number of contigs within trusted genomes.
        min_N50 : int
            Minimum N50 of trusted genomes.

        Returns
        -------
        dict : d[assembly_accession] -> [comp, cont, # contigs, N50]
            Genome statistics for trusted genomes.
        """

        trusted_genome_stats = {}

        csv_reader = csv.reader(open(metadata_file, 'rt'))
        bHeader = True
        for row in csv_reader:
            if bHeader:
                headers = row
                genome_index = headers.index('genome')
                comp_index = headers.index('checkm_completeness')
                cont_index = headers.index('checkm_contamination')
                num_contigs_index = headers.index('contig_count')
                n50_index = headers.index('n50_contigs')
                bHeader = False
            else:
                genome_id = row[genome_index]
                comp = float(row[comp_index])
                cont = float(row[cont_index])
                num_contigs = int(row[num_contigs_index])
                N50 = float(row[n50_index])

                if (comp >= trusted_comp and
                    cont <= trusted_cont and
                    num_contigs <= max_contigs and
                    N50 >= min_N50):
                        trusted_genome_stats[genome_id] = [comp, cont, num_contigs, N50]

        return trusted_genome_stats

    def run(self, metadata_file,
                    trusted_comp,
                    trusted_cont,
                    max_contigs,
                    min_N50,
                    refseq_rep,
                    output_file):
        """Determine trusted genomes based on genome statistics.

        Parameters
        ----------
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        trusted_comp : float [0, 100]
            Minimum completeness to trust genome for marker set inference.
        trusted_cont : float [0, 100]
            Maximum contamination to trust genome for marker set inference.
        max_contigs : int
            Maximum number of contigs within trusted genomes.
        min_N50 : int
            Minimum N50 of trusted genomes.
        refseq_rep : boolean
            If true, consider only RefSeq representative and reference genomes.
        output_file : str
            Output file to contain list of trusted genomes.
        """

        representative_genomes = None
        if refseq_rep:
            _accession_to_taxid, complete_genomes, representative_genomes = ncbi.read_refseq_metadata(metadata_file, keep_db_prefix=True)

        gtdb_taxonomy = read_gtdb_taxonomy(metadata_file)
        ncbi_taxonomy = read_gtdb_ncbi_taxonomy(metadata_file)

        trusted_genomes_stats = self._trusted_genomes(metadata_file,
                                                      trusted_comp, trusted_cont,
                                                      max_contigs, min_N50)
        if representative_genomes:
            self.logger.info('Limiting genomes to RefSeq representative.')
            for genome_id in trusted_genomes_stats.keys():
                if genome_id not in representative_genomes:
                    del trusted_genomes_stats[genome_id]

        self.logger.info('Identified %d trusted genomes.' % len(trusted_genomes_stats))

        fout = open(output_file, 'w')
        fout.write('# Selection criteria:\n')
        fout.write('# Trusted completeness: %f\n' % trusted_comp)
        fout.write('# Trusted contamination: %f\n' % trusted_cont)
        fout.write('# Maximum contigs: %d\n' % max_contigs)
        fout.write('# Minimum N50: %d\n' % min_N50)
        fout.write('#\n')
        fout.write('# Genome Id\tCompleteness,Contamination,Contig count,N50\tGTDB Taxonomy\tNCBI Taxonomy\n')

        for assembly_accession, stats in trusted_genomes_stats.iteritems():
            fout.write('%s\t%s\t%s\t%s\n' % (assembly_accession,
                                                 ','.join(map(str, stats)),
                                                 ';'.join(gtdb_taxonomy.get(assembly_accession, ['none'])),
                                                 ';'.join(ncbi_taxonomy.get(assembly_accession, ['none']))))

        fout.close()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import csv

"""Functions for working with NCBI genomes and metadata."""


def read_genome_dir(genome_dir_file):
    """Parse genome dir file.

    """

    genome_dirs = {}
    for line in open(genome_dir_file):
        line_split = line.rstrip().split('\t')
        genome_dirs[line_split[0]] = line_split[1]

    return genome_dirs


def read_refseq_metadata(metadata_file, keep_db_prefix=False):
    """Parse metadata for RefSeq genomes from GTDB metadata file.

    Parameters
    ----------
    metadata_file : str
        File specifying metadata for all genomes.

    Returns
    -------
    dict : d[assembly_accession] -> taxonomy id
        Taxonomy id of assemblies.
    set
        Set of complete genomes based on NCBI assembly level.
    set
        Set of reference or representative genomes based on NCBI RefSeq category.
    """

    accession_to_taxid = {}
    complete_genomes = set()
    representative_genomes = set()

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            genome_index = row.index('genome')
            taxid_index = row.index('ncbi_taxid')
            assembly_level_index = row.index('ncbi_assembly_level')
            refseq_category_index = row.index('ncbi_refseq_category')
            bHeader = False
        else:
            assembly_accession = row[genome_index]
            if assembly_accession.startswith('RS_'):
                if not keep_db_prefix:
                    assembly_accession = assembly_accession.replace('RS_', '').replace('GB_', '')

                accession_to_taxid[assembly_accession] = row[taxid_index]

                assembly_level = row[assembly_level_index].lower()
                if assembly_level == 'complete genome':
                    complete_genomes.add(assembly_accession)

                refseq_category = row[refseq_category_index].lower()
                if 'reference' in refseq_category or 'representative' in refseq_category:
                    representative_genomes.add(assembly_accession)

    return accession_to_taxid, complete_genomes, representative_genomes




from distutils.core import setup

import os


def version():
    setupDir = os.path.dirname(os.path.realpath(__file__))
    versionFile = open(os.path.join(setupDir, 'genometreetk', 'VERSION'))
    return versionFile.read().strip()

setup(
    name='GenomeTreeTk',
    version=version(),
    author='Donovan Parks',
    author_email='donovan.parks@gmail.com',
    packages=['genometreetk', 'genometreetk.markers'],
    scripts=['bin/genometreetk'],
    package_data={'genometreetk' : ['VERSION'], '': ['distributions/*.txt']},
    url='http://pypi.python.org/pypi/genometreetk/',
    license='GPL3',
    description='A toolbox for working with genome trees.',
    long_description=open('README.md').read(),
    install_requires=[
        "numpy >= 1.8.0",
        "biolib >= 0.0.11",
        "dendropy >= 4.0.0"],
)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import operator
from collections import defaultdict

import biolib.seq_io as seq_io

from genometreetk.exceptions import GenomeTreeTkError
from genometreetk.common import (read_gtdb_genome_quality,
                                    read_gtdb_metadata,
                                    read_gtdb_taxonomy,
                                    read_gtdb_ncbi_taxonomy,
                                    read_gtdb_ncbi_organism_name,
                                    species_label)
from genometreetk.aai import aai_test, mismatches


class Representatives(object):
    """Identify representative genomes.

    Representative genomes are identified in a
    greedy manner using an amino acid identity (AAI)
    criteria.

    To ensure good representatives are selected, genomes
    are order before processing. Genomes are order first
    based on their source: RefSeq, GenBank, User. Within
    each source, genomes are order by genome quality
    (completeness - contamination). A threshold is used
    to limit representative to genomes of sufficient
    quality. Furthermore, a genome is not clustered
    to an existing representative if they have different
    valid species names. NCBI genomes are also never assigned
    to User representatives.
    """

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger()

    def _order_genomes(self, genomes_to_consider, genome_quality, trusted_user_genomes):
        """Order genomes by source and genome quality.

        Parameters
        ----------
        genomes_to_consider : list
            Genomes to order.
        genome_quality : d[genome_id] -> genome quality
            Estimate quality (completeness - contamination) of each genome.
        trusted_user_genomes : set
            Trusted User genomes to treat as if they were in GenBank.

        Returns
        -------
        list
            Genomes order by source and quality.
        """
        # sort genomes by source repository followed by genome quality
        sorted_refseq_rep_genomes = []
        sorted_genbank_rep_genomes = []
        sorted_trusted_user_rep_genomes = []
        sorted_user_rep_genomes = []
        sorted_by_quality = sorted(genome_quality.items(), key=operator.itemgetter(1), reverse=True)
        for genome_id, _quality in sorted_by_quality:
            if genome_id not in genomes_to_consider:
                continue

            if genome_id.startswith('RS_'):
                sorted_refseq_rep_genomes.append(genome_id)
            elif genome_id.startswith('GB_'):
                sorted_genbank_rep_genomes.append(genome_id)
            elif genome_id in trusted_user_genomes:
                sorted_trusted_user_rep_genomes.append(genome_id)
            elif genome_id.startswith('U_'):
                sorted_user_rep_genomes.append(genome_id)
            else:
                self.logger.error('Unrecognized genome prefix: %s' % genome_id)
                sys.exit(-1)

        return (sorted_refseq_rep_genomes
                    + sorted_genbank_rep_genomes
                    + sorted_trusted_user_rep_genomes
                    + sorted_user_rep_genomes)

    def _greedy_representatives(self,
                                    representatives,
                                    genomes_to_process,
                                    aai_threshold,
                                    ar_seqs,
                                    bac_seqs,
                                    metadata_file,
                                    trusted_user_genomes):
        """Identify additional representative genomes in a greedy fashion.

        Parameters
        ----------
        representatives : set
            Initial set of representative genomes.
        genomes_to_process : list
            Genomes to process for identification of new representatives.
        aai_threshold : float
              AAI threshold for assigning a genome to a representative.
        ar_seqs : d[genome_id] -> alignment
            Alignment of archaeal marker genes.
        bac_seqs : d[genome_id] -> alignment
            Alignment of bacterial marker genes.
        metadata_file : str
            Metadata for all genomes, including NCBI taxonomy information.
        trusted_user_genomes : set
            Trusted User genomes to treat as if they were in GenBank.

        Returns
        -------
        set
            Representative genomes.
        """

        # read taxonomy information and determine 'best' species label for each genome
        gtdb_taxonomy = read_gtdb_taxonomy(metadata_file)
        ncbi_taxonomy = read_gtdb_ncbi_taxonomy(metadata_file)
        ncbi_organism_names = read_gtdb_ncbi_organism_name(metadata_file)
        species = species_label(gtdb_taxonomy, ncbi_taxonomy, ncbi_organism_names)

        # determine genus of all genomes and representatives
        genus = {}
        reps_from_genus = defaultdict(set)
        for genome_id, t in gtdb_taxonomy.iteritems():
            g = None
            if len(t) >= 6 and t[5] != 'g__':
                g = t[5]
            elif genome_id in species:
                g = species[genome_id].split(' ')[0][3:]

            if g:
                genus[genome_id] = g

                if genome_id in representatives:
                    reps_from_genus[g].add(genome_id)

        total_genomes = len(genomes_to_process)
        processed_genomes = 0
        while len(genomes_to_process):
            processed_genomes += 1
            sys.stdout.write('==> Processed %d of %d genomes.\r' % (processed_genomes, total_genomes))
            sys.stdout.flush()

            genome_id = genomes_to_process.pop(0)

            genome_genus = genus.get(genome_id, None)
            genome_species = species.get(genome_id, None)
            genome_bac_seq = bac_seqs[genome_id]
            genome_ar_seq = ar_seqs[genome_id]

            # speed up computation by comparing genome to representatives from the same genus
            cur_reps_from_genus = reps_from_genus.get(genome_genus, set())
            bCluster = False
            for rep_id in cur_reps_from_genus:
                # do not cluster genomes from different named species
                rep_species = species.get(rep_id, None)
                if rep_species and genome_species and rep_species != genome_species:
                    continue

                # do not cluster NCBI or trusted User genomes with User representatives
                if rep_id.startswith('U_') and (not genome_id.startswith('U__') or genome_id in trusted_user_genomes):
                    continue

                rep_bac_seq = bac_seqs[rep_id]
                rep_ar_seq = ar_seqs[rep_id]

                bCluster = (aai_test(rep_bac_seq, genome_bac_seq, aai_threshold)
                            or aai_test(rep_ar_seq, genome_ar_seq, aai_threshold))
                if bCluster:
                    break

            if not bCluster:
                # compare genome to remaining representatives
                remaining_reps = representatives.difference(cur_reps_from_genus)
                for rep_id in remaining_reps:
                    # do not cluster genomes from different named species
                    rep_species = species.get(rep_id, None)
                    if rep_species and genome_species and rep_species != genome_species:
                        continue

                    # do not cluster genomes from different named groups
                    skip = False
                    for rep_taxon, taxon in zip(gtdb_taxonomy[rep_id], gtdb_taxonomy[genome_id]):
                        rep_taxon = rep_taxon[3:]
                        taxon = taxon[3:]
                        if rep_taxon and taxon and rep_taxon != taxon:
                            skip = True
                            break
                    if skip:
                        continue

                    # do not cluster NCBI or trusted User genomes with User representatives
                    if rep_id.startswith('U_') and (not genome_id.startswith('U__') or genome_id in trusted_user_genomes):
                        continue

                    rep_bac_seq = bac_seqs[rep_id]
                    rep_ar_seq = ar_seqs[rep_id]

                    bCluster = (aai_test(rep_bac_seq, genome_bac_seq, aai_threshold)
                                or aai_test(rep_ar_seq, genome_ar_seq, aai_threshold))
                    if bCluster:
                        break

            if not bCluster:
                # genome was not assigned to an existing representative,
                # so make it a new representative genome
                representatives.add(genome_id)
                if genome_genus:
                    reps_from_genus[genome_genus].add(genome_id)

        sys.stdout.write('\n')

        return representatives

    def run(self,
            refseq_representatives,
            trusted_user_file,
            ar_msa_file,
            bac_msa_file,
            aai_threshold,
            min_rep_comp,
            max_rep_cont,
            min_quality,
            max_contigs,
            min_N50,
            max_ambiguous,
            metadata_file,
            output_file):
        """Identify additional representatives based on AAI between aligned sequences.

        Parameters
        ----------
        refseq_representatives : str
            File listing RefSeq genome identifiers as initial representatives.
        trusted_user_file : str
            File listing trusted User genomes that should be treated as if they are in GenBank.
        ar_msa_file : str
            Name of file containing canonical archaeal multiple sequence alignment.
        bac_msa_file : str
            Name of file containing canonical bacterial multiple sequence alignment.
        aai_threshold : float
              AAI threshold for clustering genomes to a representative.
        min_rep_comp : float [0, 100]
            Minimum completeness for a genome to be a representative.
        max_rep_cont : float [0, 100]
            Maximum contamination for a genome to be a representative.
        min_quality : float [0, 100]
            Minimum quality (comp - 4*cont) for a genome to be a representative.
        max_contigs : int
            Maximum number of contigs for a genome to be a representative.
        min_N50 : int
            Minimum N50 for a genome to be a representative.
        max_ambiguous : int
            Maximum number of ambiguous bases for a genome to be a representative.
        metadata_file : str
            Metadata, including CheckM estimates, for all genomes.
        output_file : str
            Output file containing all genomes identified as representatives.
        """

        # read sequences
        ar_seqs = seq_io.read_fasta(ar_msa_file)
        bac_seqs = seq_io.read_fasta(bac_msa_file)
        self.logger.info('Identified %d archaeal sequences in MSA.' % len(ar_seqs))
        self.logger.info('Identified %d bacterial sequences in MSA.' % len(bac_seqs))

        if len(ar_seqs) != len(bac_seqs):
            self.logger.error('Archaeal and bacterial MSA files do not contain the same number of sequences.')
            raise GenomeTreeTkError('Error with MSA input files.')

        genome_to_consider = set(ar_seqs.keys())

        # read initial representatives
        refseq_rep_genomes = set()
        for line in open(refseq_representatives):
            if line[0] == '#':
                continue

            genome_id = line.rstrip().split('\t')[0]
            if genome_id not in genome_to_consider:
                self.logger.error('Representative genome %s has no sequence data.' % genome_id)
                sys.exit(-1)
            refseq_rep_genomes.add(genome_id)

        self.logger.info('Identified %d initial representatives.' % len(refseq_rep_genomes))
        
        # read trusted User genomes
        trusted_user_genomes = set()
        if trusted_user_file:
            for line in open(trusted_user_file):
                line_split = line.split('\t')
                trusted_user_genomes.add(line_split[0])
                
        self.logger.info('Identified %d trusted User genomes.' % len(trusted_user_genomes))        

        # read genome quality
        genome_quality = read_gtdb_genome_quality(metadata_file)

        missing_quality = genome_to_consider - set(genome_quality.keys())
        if missing_quality:
            self.logger.error('There are %d genomes with sequence data, but no genome quality information.' % len(missing_quality))
            sys.exit(-1)

        # remove existing representative genomes and genomes
        # of insufficient quality to be a representative
        genome_stats = read_gtdb_metadata(metadata_file, ['contig_count',
                                                            'n50_contigs',
                                                            'ambiguous_bases'])
        potential_reps = set()
        for genome_id in (genome_to_consider - refseq_rep_genomes):
            comp, cont, qual = genome_quality.get(genome_id, [-1, -1, -1])
            stats = genome_stats[genome_id]
            if (comp >= min_rep_comp 
                and cont <= max_rep_cont 
                and (comp - 4*cont) >= min_quality
                and stats.contig_count <= max_contigs
                and stats.n50_contigs >= min_N50
                and stats.ambiguous_bases <= max_ambiguous):
                    potential_reps.add(genome_id)

        # perform greedy identification of new representatives
        ordered_genomes = self._order_genomes(potential_reps, genome_quality, trusted_user_genomes)
        self.logger.info('Comparing %d genomes to %d representatives with threshold = %.3f.' % (len(ordered_genomes),
                                                                                                len(refseq_rep_genomes),
                                                                                                aai_threshold))
        representatives = self._greedy_representatives(refseq_rep_genomes,
                                                            ordered_genomes,
                                                            aai_threshold,
                                                            ar_seqs,
                                                            bac_seqs,
                                                            metadata_file,
                                                            trusted_user_genomes)

        self.logger.info('Identified %d representatives.' % len(representatives))

        # write out representative genomes
        fout = open(output_file, 'w')
        for rep in representatives:
            fout.write(rep + '\n')

        fout.close()
